from anaconda_auth.config import (
    _OLD_OIDC_REQUEST_HEADERS as OIDC_REQUEST_HEADERS,  # noqa: F401
)
from anaconda_auth.config import AnacondaCloudConfig  # noqa: F401
from anaconda_cloud_auth import warn

warn()



from __future__ import annotations

import os
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import IO
from typing import Any
from typing import Generator
from typing import Mapping
from typing import Protocol
from typing import Sequence
from typing import cast

import pytest
import typer
from click.testing import Result
from dotenv import load_dotenv
from keyring.backend import KeyringBackend
from pytest import MonkeyPatch
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from anaconda_auth.client import BaseClient
from anaconda_auth.token import TokenInfo
from anaconda_cli_base.cli import app

load_dotenv()


class MockedKeyring(KeyringBackend):
    """A high-priority in-memory keyring backend for testing"""

    priority = 10000.0  # type: ignore
    _data: dict = defaultdict(dict)

    def __init__(self) -> None:
        super().__init__()

    def set_password(self, service: str, username: str, password: str) -> None:
        self._data[service][username] = password

    def get_password(self, service: str, username: str) -> str | None:
        password = self._data.get(service, {}).get(username, None)
        return password

    def delete_password(self, service: str, username: str) -> None:
        _ = self._data.get(service, {}).pop(username)


@pytest.fixture(autouse=True)
def clear_mocked_keyring() -> None:
    MockedKeyring._data = defaultdict(dict)


@pytest.fixture(autouse=True)
def set_keyring_name(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.token.KEYRING_NAME", "Anaconda Test")


@pytest.fixture
def outdated_api_key() -> str:
    # This is an old token from the dev system that will always be out-of-date
    api_key = (
        "eyJhbGciOiJSUzI1NiIsImtpZCI6ImRlZmF1bHQiLCJ0eXAiOiJKV1QifQ"
        ".eyJleHAiOjE2ODkwODg3ODYsInN1YiI6ImQwNGMzNTZiLWFmZDItNGIzZ"
        "S04MGYyLTQwMzExM2UwOTc0YiJ9.tTi_gttpQWhiTy_Uh0bDohN34mqd_6"
        "AHvyXf8_R5PFxjI-z9Ei0S3XCm9siP0RfyJx2j08SRs3FwXpkT8b8jP__C"
        "h-Y4K-zXYksZnTGcQ77YhKQCoKpGSpGlE4yD6gRXRRDT7bHs4H7gf4e6iD"
        "1Vdcq0yx5-5h-CbBgSwS9LSpJ_HDZBUy-xbRrw0aD36aQ5qs6huswgCOQa"
        "9YrYfsrSbZW8uY48LAt5Y69t8x1twNBI5_Cumx-JEZuDLQxq7HQp7wKldE"
        "tbycV5uemKjyR1Qeuva2zCKYB3FEXdTEiWHhTzhSQ-3-xjUrIZvpfGJd3G"
        "CzXlkUhpeDoj2KbSN-Lq0Q"
    )
    return api_key


@pytest.fixture
def outdated_token_info(outdated_api_key: str) -> TokenInfo:
    return TokenInfo(api_key=outdated_api_key, domain="mocked-domain")


@pytest.fixture()
def tmp_cwd(monkeypatch: MonkeyPatch, tmp_path: Path) -> Path:
    """Create & return a temporary directory after setting current working directory to it."""
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def is_not_none() -> Any:
    """
    An object that can be used to test whether another is None.

    This is particularly useful when testing contents of collections, e.g.:

    ```python
    def test_data(data, is_not_none):
        assert data == {"some_key": is_not_none, "some_other_key": 5}
    ```

    """

    class _NotNone:
        def __eq__(self, other: Any) -> bool:
            return other is not None

    return _NotNone()


@pytest.fixture
def disable_dot_env(mocker: MockerFixture) -> None:
    from anaconda_auth.config import AnacondaAuthConfig

    mocker.patch.dict(AnacondaAuthConfig.model_config, {"env_file": ""})


@pytest.fixture(autouse=True)
def disable_config_toml(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.setenv("ANACONDA_CONFIG_TOML", str(tmp_path / "empty-config.toml"))


@pytest.fixture
def api_key() -> str | None:
    return os.getenv("TEST_API_KEY")


@pytest.fixture()
def integration_test_client(api_key: str | None) -> BaseClient:
    c = BaseClient(api_key=api_key)
    return c


@pytest.fixture()
def save_api_key_to_token(api_key: str | None) -> Generator[None, None, None]:
    from anaconda_auth.config import AnacondaAuthConfig

    conf = AnacondaAuthConfig()
    token = TokenInfo(api_key=api_key, domain=conf.domain)
    token.save()
    yield


def pytest_addoption(parser):  # type: ignore
    """Defines custom CLI options."""
    parser.addoption(
        "--integration",
        action="store_true",
        dest="integration",
        default=False,
        help="enable integration tests",
    )


def pytest_collection_modifyitems(config, items):  # type: ignore
    """Auto-mark each test in the integration directory, and enable them based on CLI flag."""
    integration_test_root_dir = Path(__file__).parent / "integration"
    run_integration_tests = config.getoption("--integration")
    for item in items:
        # Here, we add a marker to any test in the "tests/integration" directory
        if integration_test_root_dir in Path(item.fspath).parents:
            item.add_marker(pytest.mark.integration)

        # Add a skip marker if the CLI option is not used. We use an additional marker so that we can
        # independently select integrations with `pytest -m integration` and enable them with `--integration`.
        if "integration" in item.keywords and not run_integration_tests:
            item.add_marker(pytest.mark.skip(reason="need --integration to run"))


@pytest.fixture
def with_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.config.AnacondaAuthConfig.aau_token", "anon-token")


@pytest.fixture
def without_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.config.AnacondaAuthConfig.aau_token", None)


class MockResponse:
    def __init__(
        self,
        *,
        status_code: int,
        json_data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ):
        self.status_code = status_code
        self.json_data = json_data
        self.headers = headers or {}

    def json(self) -> dict[str, Any]:
        return self.json_data or {}


class MockedRequest:
    def __init__(
        self,
        *,
        response_status_code: int,
        response_data: dict[str, Any] | None = None,
        response_headers: dict[str, str] | None = None,
    ) -> None:
        self.response_status_code = response_status_code
        self.response_data = response_data
        self.response_headers = response_headers
        self.called_with_args: tuple[Any] = ()  # type: ignore
        self.called_with_kwargs: dict[str, Any] = {}

    def __call__(self, *args: Any, **kwargs: Any) -> MockResponse:
        self.called_with_args = args  # type: ignore
        self.called_with_kwargs = kwargs
        return MockResponse(
            status_code=self.response_status_code,
            json_data=self.response_data,
            headers=self.response_headers,
        )


class CLIInvoker(Protocol):
    def __call__(
        self,
        # app: typer.Typer,
        args: str | Sequence[str] | None = None,
        input: bytes | str | IO[Any] | None = None,
        env: Mapping[str, str] | None = None,
        catch_exceptions: bool = True,
        color: bool = False,
        **extra: Any,
    ) -> Result: ...


@pytest.fixture()
def invoke_cli(tmp_cwd: Path) -> CLIInvoker:
    """Returns a function, which can be used to call the CLI from within a temporary directory."""

    runner = CliRunner()

    return partial(runner.invoke, cast(typer.Typer, app))



import logging
import uuid
import warnings
import webbrowser
from typing import Optional
from urllib.parse import urlencode

import pkce
import requests

from anaconda_auth import __version__
from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.exceptions import AuthenticationError
from anaconda_auth.exceptions import TokenNotFoundError
from anaconda_auth.handlers import capture_auth_code
from anaconda_auth.token import TokenInfo
from anaconda_cli_base.console import console

logger = logging.getLogger(__name__)


def make_auth_code_request_url(
    code_challenge: str, state: str, config: Optional[AnacondaAuthConfig] = None
) -> str:
    """Build the authorization code request URL."""

    if config is None:
        config = AnacondaAuthConfig()

    authorization_endpoint = config.oidc.authorization_endpoint
    client_id = config.client_id
    redirect_uri = config.redirect_uri

    params = dict(
        client_id=client_id,
        response_type="code",
        scope="openid email profile offline_access",
        state=state,
        redirect_uri=redirect_uri,
        code_challenge=code_challenge,
        code_challenge_method="S256",
    )
    encoded_params = urlencode(params)
    url = f"{authorization_endpoint}?{encoded_params}"

    return url


def _send_auth_code_request(
    code_challenge: str, state: str, config: AnacondaAuthConfig
) -> None:
    """Open the authentication flow in the browser."""
    url = make_auth_code_request_url(code_challenge, state, config)
    webbrowser.open(url)


def refresh_access_token(refresh_token: str, config: AnacondaAuthConfig) -> str:
    """Refresh and save the tokens."""
    response = requests.post(
        config.oidc.token_endpoint,
        data={
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": config.client_id,
        },
        verify=config.ssl_verify,
    )
    response.raise_for_status()
    response_data = response.json()

    access_token = response_data["access_token"]
    return access_token


def request_access_token(
    auth_code: str, code_verifier: str, config: AnacondaAuthConfig
) -> str:
    """Request an access token using the provided authorization code and code verifier."""
    token_endpoint = config.oidc.token_endpoint
    client_id = config.client_id
    redirect_uri = config.redirect_uri

    response = requests.post(
        token_endpoint,
        data=dict(
            grant_type="authorization_code",
            client_id=client_id,
            code=auth_code,
            redirect_uri=redirect_uri,
            code_verifier=code_verifier,
        ),
        verify=config.ssl_verify,
    )
    result = response.json()

    if "error" in result:
        raise AuthenticationError(
            f"Error getting JWT: {result.get('error')} - {result.get('error_description')}"
        )

    access_token = result.get("access_token")
    return access_token


def _do_auth_flow(config: Optional[AnacondaAuthConfig] = None) -> str:
    """Do the browser-based auth flow and return the short-lived access_token and id_token tuple."""
    config = config or AnacondaAuthConfig()

    state = str(uuid.uuid4())
    code_verifier, code_challenge = pkce.generate_pkce_pair(code_verifier_length=128)

    _send_auth_code_request(code_challenge, state, config)

    # Listen for the response
    auth_code = capture_auth_code(config.redirect_uri, state=state, config=config)
    logger.debug("Authentication successful! Getting JWT token.")

    # Do auth code exchange
    return request_access_token(auth_code, code_verifier, config)


def _login_with_username(config: Optional[AnacondaAuthConfig] = None) -> str:
    """Prompt for username and password and log in with the password grant flow."""
    warnings.warn(
        "Basic login with username/password is deprecated and will be disabled soon.",
        UserWarning,
        stacklevel=0,
    )

    if config is None:
        config = AnacondaAuthConfig()

    username = console.input("Please enter your email: ")
    password = console.input("Please enter your password: ", password=True)
    response = requests.post(
        config.oidc.token_endpoint,
        data={
            "grant_type": "password",
            "username": username,
            "password": password,
        },
        verify=config.ssl_verify,
    )
    response_data = response.json()
    response.raise_for_status()

    access_token = response_data["access_token"]
    return access_token


def _do_login(config: AnacondaAuthConfig, basic: bool) -> None:
    if basic:
        access_token = _login_with_username(config=config)
    else:
        access_token = _do_auth_flow(config=config)
    api_key = get_api_key(access_token, config.ssl_verify)
    token_info = TokenInfo(api_key=api_key, domain=config.domain)
    token_info.save()


def get_api_key(
    access_token: str, ssl_verify: bool = True, config: Optional[AnacondaAuthConfig] = None
) -> str:
    config = config or AnacondaAuthConfig()

    headers = {"Authorization": f"Bearer {access_token}"}

    aau_token = config.aau_token
    if aau_token is not None:
        headers["X-AAU-CLIENT"] = aau_token

    # Retry logic until we stabilize on new API
    urls = [
        f"https://{config.auth_domain}/api/auth/api-keys",
        f"https://{config.domain}/api/iam/api-keys",
    ]
    for url in urls:
        response = requests.post(
            url,
            json=dict(
                scopes=["cloud:read", "cloud:write", "repo:read"],
                tags=[f"anaconda-auth/v{__version__}"],
            ),
            headers=headers,
            verify=ssl_verify,
        )
        if response.status_code == 201:
            break
    else:
        console.print("Error retrieving an API key")
        raise AuthenticationError
    return response.json()["api_key"]


def _api_key_is_valid(config: AnacondaAuthConfig) -> bool:
    try:
        valid = not TokenInfo.load(config.domain).expired
    except TokenNotFoundError:
        valid = False

    return valid


def login(
    config: Optional[AnacondaAuthConfig] = None,
    basic: bool = False,
    force: bool = False,
    ssl_verify: bool = True,
) -> None:
    """Log into anaconda.com and store the token information in the keyring."""
    if config is None:
        config = AnacondaAuthConfig(ssl_verify=ssl_verify)

    if force or not _api_key_is_valid(config=config):
        _do_login(config=config, basic=basic)


def logout(config: Optional[AnacondaAuthConfig] = None) -> None:
    """Log out of anaconda.com."""
    if config is None:
        config = AnacondaAuthConfig()

    try:
        token_info = TokenInfo.load(domain=config.domain)
        token_info.delete()
    except TokenNotFoundError:
        pass

    if config.domain != "anaconda.com":
        # Since anaconda.com is the default, don't do anything special if
        # User explicitly overrode the configured domain.
        return

    # If the request was for anaconda.com (the default), also remove
    # anaconda.cloud if it exists. This is just an edge case for the
    # likely rare scenario where a user has a stored token for both
    # domains.
    try:
        token_info = TokenInfo.load(domain="anaconda.cloud")
        token_info.delete()
    except TokenNotFoundError:
        pass


def is_logged_in() -> bool:
    config = AnacondaAuthConfig()
    try:
        token_info = TokenInfo.load(domain=config.domain)
    except TokenNotFoundError:
        token_info = None

    return token_info is not None



import pytest
import requests
from pytest import MonkeyPatch
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker

from anaconda_auth.config import AnacondaAuthConfig


@pytest.fixture(autouse=True)
def mock_openid_configuration(requests_mock: RequestMocker):
    config = AnacondaAuthConfig()
    """Mock return value of openid configuration to prevent requiring actual network calls."""
    expected = {
        "authorization_endpoint": f"https://auth.{config.domain}/api/auth/oauth2/authorize",
        "token_endpoint": f"https://auth.{config.domain}/api/auth/oauth2/token",
    }
    requests_mock.get(url=config.well_known_url, json=expected)


def test_well_known_headers(mocker: MockerFixture) -> None:
    spy = mocker.spy(requests, "get")

    config = AnacondaAuthConfig()
    assert config.oidc
    spy.assert_called_once()
    assert (
        spy.call_args.kwargs.get("headers", {})
        .get("User-Agent")
        .startswith("anaconda-auth")
    )


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_env_variable_over_default(monkeypatch: MonkeyPatch, prefix: str) -> None:
    monkeypatch.setenv(f"{prefix}_DOMAIN", "set-in-env")
    config = AnacondaAuthConfig()
    assert config.domain == "set-in-env"


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_init_arg_over_env_variable(monkeypatch: MonkeyPatch, prefix: str) -> None:
    monkeypatch.setenv(f"{prefix}_DOMAIN", "set-in-env")
    config = AnacondaAuthConfig(domain="set-in-init")
    assert config.domain == "set-in-init"


def test_override_auth_domain_env_variable(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv(
        "ANACONDA_AUTH_AUTH_DOMAIN_OVERRIDE", "another-auth.anaconda.com"
    )
    config = AnacondaAuthConfig()
    assert config.auth_domain == "another-auth.anaconda.com"



# The plugin entrypoints are loaded via this module as specified in pyproject.toml
# We do this to be very conservative and catch potential import errors without surfacing to users.
# Plugins were first introduced in conda=22.11.0
# The ChannelAuthBase class was introduced in conda=23.7.0
try:
    from anaconda_auth._conda.plugins import *  # noqa: F403
except ImportError:
    pass



from __future__ import annotations

from datetime import datetime
from uuid import UUID

import typer
from pydantic import BaseModel
from rich.prompt import Confirm
from rich.table import Table

from anaconda_auth.actions import _do_auth_flow
from anaconda_auth.client import BaseClient
from anaconda_auth.token import RepoToken
from anaconda_auth.token import TokenInfo
from anaconda_cli_base import console
from anaconda_cli_base.console import select_from_list

app = typer.Typer(name="token")


class TokenInfoResponse(BaseModel):
    id: UUID
    expires_at: datetime


class TokenCreateResponse(BaseModel):
    token: str
    expires_at: datetime


class OrganizationData(BaseModel):
    id: UUID
    name: str
    title: str


class SubscriptionData(BaseModel):
    org_id: UUID
    product_code: str


class RepoAPIClient(BaseClient):
    def __init__(self) -> None:
        super().__init__()
        self._access_token: str | None = None

    def _ensure_access_token(self) -> None:
        """Some endpoints do not accept API keys, so this method ensures we perform
        an interactive authentication and then cache the access token.
        """
        if self._access_token is not None:
            return
        self._access_token = _do_auth_flow()
        self.auth.api_key = self._access_token  # type: ignore

    def _get_repo_token_info(self, org_name: str) -> TokenInfoResponse | None:
        """Return the token information, if it exists.

        Args:
            org_name: The name of the organization.

        Returns:
            The token information, including its id and expiration date, or
            None if a token doesn't exist.
        """
        self._ensure_access_token()
        response = self.get(
            f"/api/organizations/{org_name}/ce/current-token",
        )
        if response.status_code == 404:
            return None
        response.raise_for_status()
        return TokenInfoResponse(**response.json())

    def _create_repo_token(self, org_name: str) -> TokenCreateResponse:
        """Create a new repo token.

        Args:
            org_name: The name of the organization.

        Returns:
            The token information, including its value and expiration date.
        """
        self._ensure_access_token()
        response = self.put(
            f"/api/organizations/{org_name}/ce/current-token",
            json={"confirm": "yes"},
        )
        return TokenCreateResponse(**response.json())

    def issue_new_token(self, org_name: str) -> str:
        """Issue a new repository token from anaconda.com."""
        existing_token_info = self._get_repo_token_info(org_name=org_name)

        if existing_token_info is not None:
            console.print(
                f"An existing token already exists for the organization [cyan]{org_name}[/cyan]."
            )
            console.print(
                "Reissuing a new token will revoke and deactivate any existing token access. This action can't be undone."
            )
            should_continue = Confirm.ask("Proceed?", default=False)
            if not should_continue:
                raise typer.Abort()

        response = self._create_repo_token(org_name=org_name)

        console.print(
            f"Your conda token has been installed and expires [cyan]{response.expires_at}[/cyan]. To view your token(s), you can use [cyan]anaconda token list[/cyan]"
        )
        return response.token

    def get_organizations_for_user(self) -> list[OrganizationData]:
        """Get a list of all organizations the user belongs to."""
        response = self.get("/api/organizations/my")
        response.raise_for_status()
        data = response.json()
        return [OrganizationData(**item) for item in data]

    def get_business_organizations_for_user(self) -> list[OrganizationData]:
        """Get a list of all organizations the user belongs to that have a Business subscription."""
        organizations = self.get_organizations_for_user()
        subscriptions = [
            SubscriptionData(**sub) for sub in self.account.get("subscriptions", [])
        ]
        business_subscription_org_ids = [
            sub.org_id for sub in subscriptions if "starter" not in sub.product_code
        ]
        return [org for org in organizations if org.id in business_subscription_org_ids]


def _print_repo_token_table(
    tokens: list[RepoToken], legacy_tokens: dict[str, str]
) -> None:
    table = Table(title="Anaconda Repository Tokens", title_style="green")

    table.add_column("Organization")
    table.add_column("Channel URL")
    table.add_column("Token")

    from anaconda_auth._conda.repo_config import REPO_URL

    for repo_token in tokens:
        channel_url = f"{REPO_URL}{repo_token.org_name}/*"
        table.add_row(repo_token.org_name, channel_url, repo_token.token)

    for url, token in legacy_tokens.items():
        table.add_row(None, url, token)

    console.print(table)


def _select_org_name(client: RepoAPIClient) -> str:
    organizations = client.get_business_organizations_for_user()

    if not organizations:
        console.print("No organizations found.")
        raise typer.Abort()

    if len(organizations) == 1:
        org_name = organizations[0].name
        console.print(
            f"Only one organization found, automatically selecting: {org_name}"
        )
        return org_name

    name_map = {}
    choices = []
    for org in organizations:
        key = f"{org.title} ([cyan]{org.name}[/cyan])"
        name_map[key] = org.name
        choices.append(key)

    org_title = select_from_list(
        "Please select an organization:",
        choices=choices,
    )
    return name_map[org_title]


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main() -> None:
    """Manage your Anaconda repo tokens."""


@app.command(name="list")
def list_tokens() -> None:
    """List all installed repository tokens."""
    from anaconda_auth._conda.repo_config import token_list

    tokens = token_list()

    token_info = TokenInfo.load(create=True)
    repo_tokens = token_info.repo_tokens

    if not (tokens or repo_tokens):
        console.print(
            "No repo tokens are installed. Run [cyan]anaconda token install[/cyan]."
        )
        raise typer.Abort()

    _print_repo_token_table(tokens=repo_tokens, legacy_tokens=tokens)


@app.command(name="install")
def install_token(
    token: str = typer.Argument(
        "", help="Optionally, provide the token received via email or web interface."
    ),
    org_name: str = typer.Option("", "-o", "--org", help="Organization name (slug)."),
    set_default_channels: bool = typer.Option(
        True, help="Automatically configure default channels."
    ),
) -> None:
    """Create and install a new repository token."""
    client = RepoAPIClient()

    if not org_name:
        org_name = _select_org_name(client)

    if not token:
        token = client.issue_new_token(org_name=org_name)

    from anaconda_auth._conda import repo_config

    try:
        repo_config.validate_token(token, no_ssl_verify=False)
    except repo_config.CondaTokenError as e:
        raise typer.Abort(e)

    token_info = TokenInfo.load(create=True)
    token_info.set_repo_token(org_name, token)
    token_info.save()

    msg = "Your token has been installed and validated"

    repo_config.configure_plugin()

    if set_default_channels:
        repo_config.configure_default_channels()
        msg += ", and conda has been configured"

    console.print(f"Success! {msg}.")


@app.command(name="config")
def configure_conda(
    force: bool = typer.Option(
        False, help="Force configuration of default channels without prompt."
    ),
) -> None:
    """Configure conda's default channels to access Anaconda's premium repository."""
    from anaconda_auth._conda import repo_config

    repo_config.configure_plugin()
    repo_config.configure_default_channels(force=force)


@app.command(name="uninstall")
def uninstall_token(org_name: str = typer.Option("", "-o", "--org")) -> None:
    """Uninstall a repository token for a specific organization."""
    # TODO: Add --all option
    if not org_name:
        # TODO: We should try to load this dynamically and present a picker
        console.print("Must explicitly provide an [cyan]--org[/cyan] option")
        raise typer.Abort()

    token_info = TokenInfo.load()
    token_info.delete_repo_token(org_name=org_name)
    token_info.save()

    console.print(
        f"Successfully deleted token for organization: [cyan]{org_name}[/cyan]"
    )



"""
Configure Conda to use Anaconda Commercial Edition.
"""

from __future__ import annotations

import contextlib
import io
import json
import os
import sys
import warnings
from os.path import abspath
from os.path import expanduser
from os.path import join
from typing import Any
from urllib.parse import urljoin

import conda
import conda.gateways.logging  # noqa: F401
from conda.base.context import context
from conda.base.context import reset_context
from conda.cli import main as run_command
from conda.exceptions import CondaKeyError
from conda.gateways.anaconda_client import read_binstar_tokens
from conda.gateways.anaconda_client import remove_binstar_token
from conda.gateways.anaconda_client import set_binstar_token
from conda.gateways.connection.session import CondaSession
from conda.models.channel import Channel
from packaging import version
from rich.prompt import Confirm

from anaconda_auth._conda.condarc import CondaRC
from anaconda_cli_base import console

CONDA_VERSION = version.parse(conda.__version__)

REPO_URL = os.getenv("CONDA_TOKEN_REPO_URL", "https://repo.anaconda.cloud/repo/")
MAIN_CHANNEL = "main"
ACTIVE_CHANNELS = ["r", "msys2"]
ARCHIVE_CHANNELS = ["free", "mro-archive", "pro"]

user_rc_path = abspath(expanduser("~/.condarc"))
escaped_user_rc_path = user_rc_path.replace("%", "%%")
escaped_sys_rc_path = abspath(join(sys.prefix, ".condarc")).replace("%", "%%")


class Commands:
    """Names for conda commands used."""

    CONFIG = "config"
    CLEAN = "clean"


class CondaTokenError(RuntimeError):
    pass


class CondaVersionWarning(UserWarning):
    pass


def can_restore_free_channel() -> bool:
    return CONDA_VERSION >= version.parse("4.7.0")


def get_ssl_verify() -> bool:
    context = reset_context()
    return context.ssl_verify


def clean_index() -> None:
    """Runs conda clean -i.

    It is important to remove index cache when
    changing the condarc to ensure that the downloaded
    repodata is correct.
    """
    run_command(Commands.CLEAN, "-i", "-y", "-q")


def validate_token(token: str, no_ssl_verify: bool = False) -> None:
    """Checks that token can be used with the repository."""

    # Force ssl_verify: false
    if no_ssl_verify:
        context.ssl_verify = False  # type: ignore

    # Use CondaSession to be compatible with ssl_verify: truststore
    # https://conda.io/projects/conda/en/latest/user-guide/configuration/settings.html#ssl-verify-ssl-verification
    # Clear metaclass cache to create new session checking ssl_verify
    if hasattr(CondaSession, "cache_clear"):
        # not present in conda < January 2024
        CondaSession.cache_clear()
    else:
        # what cache_clear() does
        try:
            CondaSession._thread_local.sessions.clear()  # type: ignore
        except AttributeError:
            # AttributeError: thread's session cache has not been initialized
            pass

    session = CondaSession()

    # Ensure the index cache is cleaned first
    clean_index()

    channel = Channel(urljoin(REPO_URL, "main/noarch/repodata.json"))
    channel.token = token
    token_url = str(channel.url(with_credentials=True))

    r = session.head(token_url, verify=session.verify)
    if r.status_code != 200:
        raise CondaTokenError(
            "The token could not be validated. Please check that you have typed it correctly."
        )


def configure_plugin(should_set_default_channels: bool = False) -> None:
    """Configure the user's .condarc file with the anaconda-auth plugin.

    We install the auth-handler plugin by writing the "channel_settings" key and associate
    all premium repo channels with this auth handler.

    """
    condarc = CondaRC()
    condarc.backup()
    condarc.update_channel_settings(
        "https://repo.anaconda.cloud/*", "anaconda-auth", username=None
    )
    condarc.save()


def enable_extra_safety_checks(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Enable package signature verification.

    This will set extra_safety_checks: True and
    signing_metadata_url_base in the CondaRC file.
    """
    if CONDA_VERSION < version.parse("4.10.1"):
        warnings.warn(
            "You need upgrade to at least Conda version 4.10.1 to enable package signature verification.",
            CondaVersionWarning,
        )
        return

    condarc_file_args = []
    if condarc_system:
        condarc_file_args.append("--system")
    elif condarc_env:
        condarc_file_args.append("--env")
    elif condarc_file:
        condarc_file_args.append(f"--file={condarc_file}")

    safety_check_args = ["--set", "extra_safety_checks", "true"]
    safety_check_args.extend(condarc_file_args)
    run_command(Commands.CONFIG, *safety_check_args)

    metadata_url_args = ["--set", "signing_metadata_url_base", REPO_URL.rstrip("/")]
    metadata_url_args.extend(condarc_file_args)
    run_command(Commands.CONFIG, *metadata_url_args)


def disable_extra_safety_checks(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Disable package signature verification.

    This will set extra_safety_checks: false and remove
    signing_metadata_url_base in the CondaRC file.
    """

    if CONDA_VERSION < version.parse("4.10.1"):
        return

    condarc_file_args = []
    if condarc_system:
        condarc_file_args.append("--system")
    elif condarc_env:
        condarc_file_args.append("--env")
    elif condarc_file:
        condarc_file_args.append(f"--file={condarc_file}")

    safety_check_args = ["--set", "extra_safety_checks", "false"]
    safety_check_args.extend(condarc_file_args)
    try:
        run_command(Commands.CONFIG, *safety_check_args)
    except CondaKeyError:
        pass

    metadata_url_args = ["--remove-key", "signing_metadata_url_base"]
    metadata_url_args.extend(condarc_file_args)
    try:
        run_command(Commands.CONFIG, *metadata_url_args)
    except CondaKeyError:
        pass


def _set_add_anaconda_token(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Run conda config --set add_anaconda_token true.

    Setting this parameter to true ensures that the token
    is used when making requests to the repository.
    """
    config_args = ["--set", "add_anaconda_token", "true"]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    run_command(Commands.CONFIG, *config_args)


def _set_ssl_verify_false(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Run conda config --set ssl_verify false.

    Setting this parameter to false disables all
    SSL verification for conda activities
    """
    config_args = ["--set", "ssl_verify", "false"]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    run_command(Commands.CONFIG, *config_args)


def _unset_restore_free_channel(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Runs conda config --set restore_free_channel false.

    The free channel is provided by Commercial Edition as
    and should be added directly."""
    config_args = ["--set", "restore_free_channel", "false"]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    run_command(Commands.CONFIG, *config_args, use_exception_handler=True)


def _set_channel(
    channel: str,
    prepend: bool = True,
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Adds a named Commercial Edition channel to default_channels."""
    channel_url = urljoin(REPO_URL, channel)

    config_args = [
        "--prepend" if prepend else "--append",
        "default_channels",
        channel_url,
    ]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    run_command(Commands.CONFIG, *config_args)


def _get_from_condarc(
    key: str,
    default: Any = None,
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> Any:
    """Retrieve the existing default_channels from the user's `.condarc` file.

    If the user does not have a "default_channels" section, an empty list is returned.

    """
    config_args = ["--get", key, "--json"]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    # Capture the JSON output from stdout
    string_io = io.StringIO()
    with contextlib.redirect_stdout(string_io):
        run_command(Commands.CONFIG, *config_args)

    try:
        result = json.loads(string_io.getvalue())
    except json.JSONDecodeError:
        result = {}

    return result.get("get", {}).get(key, default)


def _get_default_channels(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> list[str]:
    """Retrieve the existing default_channels from the user's `.condarc` file.

    If the user does not have a "default_channels" section, an empty list is returned.

    """
    return _get_from_condarc(
        "default_channels",
        default=[],
        condarc_system=condarc_system,
        condarc_env=condarc_env,
        condarc_file=condarc_file,
    )


def _remove_default_channels(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
) -> None:
    """Runs conda config --remove-key default_channels

    It is best to remove the default_channels in case they
    are not set at the default values before configuring
    Commercial Edition.
    """
    config_args = ["--remove-key", "default_channels"]

    if condarc_system:
        config_args.append("--system")
    elif condarc_env:
        config_args.append("--env")
    elif condarc_file:
        config_args.append(f"--file={condarc_file}")

    # We suppress a CondaKeyError below to prevent logging an internal conda error
    # to the terminal.
    err_io = io.StringIO()
    with contextlib.redirect_stderr(err_io):
        try:
            run_command(Commands.CONFIG, *config_args)
        except CondaKeyError:
            pass

    err_string = err_io.getvalue()
    if "CondaKeyError" not in err_string.strip():
        print(err_string, file=sys.stderr)


def _prompt_to_set_default_channels() -> bool:
    """Prompt the user for whether they would like to set their default channels."""
    existing_default_channels = _get_default_channels()
    if existing_default_channels:
        console.print("Existing default channels found:")
        for c in existing_default_channels:
            console.print(f"- {c}")
        console.print(
            "This action will override the existing default_channels setting."
        )
    else:
        console.print("Prepared to set default channels in .condarc.")

    return Confirm.ask("Proceed?", default=False)


def configure_default_channels(
    condarc_system: bool = False,
    condarc_env: bool = False,
    condarc_file: str | None = None,
    include_archive_channels: list[str] | None = None,
    force: bool = False,
) -> None:
    """Configure the default_channels to utilize only Commercial Edition.


    This function performs the following actions
    1. unset default_channels if it exists in the condarc
    2. unset restore_free_channel if it exists in the condarc
    3. Add the main, r, and msys2 channels to default_channels
    4. Optionally add any of the archive channels:
       free, pro, mro, mro-archive
    """
    console.print("Configuring your [cyan].condarc[/cyan] file")

    existing_default_channels = _get_default_channels()

    existing_default_channels_short_names = [
        c.removeprefix(REPO_URL) for c in existing_default_channels
    ]
    if set(existing_default_channels_short_names) == set(
        [MAIN_CHANNEL] + ACTIVE_CHANNELS
    ):
        console.print("Default channels already configured, nothing to do.")
        return

    if not (force or _prompt_to_set_default_channels()):
        return

    _remove_default_channels(condarc_system, condarc_env, condarc_file)

    if can_restore_free_channel() and _get_from_condarc(
        "restore_free_channel",
        condarc_system=condarc_system,
        condarc_env=condarc_env,
        condarc_file=condarc_file,
    ):
        _unset_restore_free_channel(condarc_system, condarc_env, condarc_file)

    _set_channel(
        MAIN_CHANNEL,
        prepend=True,
        condarc_system=condarc_system,
        condarc_env=condarc_env,
        condarc_file=condarc_file,
    )

    for c in ACTIVE_CHANNELS:
        _set_channel(
            c,
            prepend=False,
            condarc_system=condarc_system,
            condarc_env=condarc_env,
            condarc_file=condarc_file,
        )

    if include_archive_channels is None:
        return

    for c in include_archive_channels:
        if c in ARCHIVE_CHANNELS:
            _set_channel(
                c,
                prepend=False,
                condarc_system=condarc_system,
                condarc_env=condarc_env,
                condarc_file=condarc_file,
            )
        else:
            raise ValueError(
                f"The archive channel {c} is not one of {', '.join(ARCHIVE_CHANNELS)}"
            )


def token_list() -> dict[str, str]:
    """Return a dictionary of tokens for all configured repository urls.

    Note that this function will return tokens configured for non-Commercial Edition
    urls."""
    return read_binstar_tokens()


def token_remove(
    system: bool = False, env: bool = False, file: str | None = None
) -> None:
    """Completely remove the Commercial Edition token and default_channels.

    This function performs three actions.
    1. Remove the token
    2. Remove the custom default_channels in the condarc
    3. Disable package signature verification
    4. Run conda clean -i
    """
    remove_binstar_token(REPO_URL)
    _remove_default_channels(system, env, file)
    disable_extra_safety_checks(system, env, file)
    clean_index()


def token_set(
    token: str,
    system: bool = False,
    env: bool = False,
    file: str | None = None,
    include_archive_channels: list[str] | None = None,
    no_ssl_verify: bool = False,
    enable_signature_verification: bool = False,
    force: bool = False,
) -> None:
    """Set the Commercial Edition token and configure default_channels.


    This function performs 4 actions.
    1. Remove previous Commercial Edition token if present.
    2. Add token.
    3. Configure default_channels in the condarc file.
    4. Optionally enable Conda package signature verification
    5. Run conda clean -i
    """
    remove_binstar_token(REPO_URL)

    set_binstar_token(REPO_URL, token)
    _set_add_anaconda_token(system, env, file)

    if no_ssl_verify:
        _set_ssl_verify_false(system, env, file)

    if enable_signature_verification:
        enable_extra_safety_checks(system, env, file)

    configure_default_channels(system, env, file, include_archive_channels, force=force)
    clean_index()



import warnings

from anaconda_auth import __version__ as __version__
from anaconda_auth import client_factory as client_factory
from anaconda_auth import login as login
from anaconda_auth import logout as logout


def warn() -> None:
    warnings.warn(
        "Please replace imports from `import anaconda_cloud_auth` to `import anaconda_auth`",
        DeprecationWarning,
    )


warn()



import base64
import datetime as dt
import json
import logging
import os
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from urllib.error import HTTPError

import jwt
import keyring
from keyring.backend import KeyringBackend
from keyring.backend import properties
from keyring.errors import PasswordDeleteError
from keyring.errors import PasswordSetError
from pydantic import BaseModel
from pydantic import Field

from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.exceptions import TokenExpiredError
from anaconda_auth.exceptions import TokenNotFoundError

# Note: we can remove this if we pin keyring>=23.9.0
try:
    classproperty = properties.classproperty
except AttributeError:
    _KeyringClassMethod = Callable[[Type[KeyringBackend]], Any]

    def classproperty(method: _KeyringClassMethod) -> _KeyringClassMethod:  # type: ignore
        return properties.ClassProperty(classmethod(method))  # type: ignore


logger = logging.getLogger(__name__)

# TODO: Rename to "Anaconda" and then migrate existing
KEYRING_NAME = "Anaconda Cloud"


# Type aliases
LocalKeyringData = Dict[str, Dict[str, str]]
OrgName = str
TokenString = str


def _as_base64_string(payload: str) -> str:
    """Encode a string to a base64 string"""
    return base64.b64encode(payload.encode("utf-8")).decode("utf-8")


class NavigatorFallback(KeyringBackend):
    priority = 0.1  # type: ignore

    @classproperty
    def viable(cls) -> bool:
        try:
            import anaconda_navigator  # noqa: F401

            return True
        except ModuleNotFoundError:
            return False

    def set_password(self, service: str, username: str, password: str) -> None:
        raise PasswordSetError("This keyring cannot set passwords")

    def _get_auth_domain(self) -> str:
        from anaconda_navigator.config import CONF as navigator_config

        known_mapping = {"https://anaconda.cloud": "id.anaconda.cloud"}

        cloud_base_url: str = navigator_config.get(
            "main", "cloud_base_url", "https://anaconda.cloud"
        ).strip("/")
        return known_mapping[cloud_base_url]

    def get_password(self, service: str, username: str) -> Union[str, None]:
        try:
            from anaconda_navigator.api.nucleus.token import NucleusToken

            auth_domain = self._get_auth_domain()
        except ImportError:
            return None

        if service != KEYRING_NAME and username != auth_domain:
            return None

        token = NucleusToken.from_file()
        if token is not None:
            from anaconda_auth.actions import get_api_key
            from anaconda_auth.actions import refresh_access_token

            config = AnacondaAuthConfig(domain=auth_domain)
            if not token.valid:
                try:
                    access_token = refresh_access_token(
                        token.refresh_token, config=config
                    )
                except HTTPError:
                    return None
            else:
                access_token = token.access_token

            api_key = get_api_key(access_token)
            token_info = {
                "username": token.username,
                "api_key": api_key,
                "domain": config.domain,
            }
            payload = json.dumps(token_info)
            encoded = _as_base64_string(payload)
            keyring.set_password(KEYRING_NAME, auth_domain, encoded)

            return encoded
        return None

    def delete_password(self, service: str, username: str) -> None:
        auth_domain = self._get_auth_domain()
        try:
            from anaconda_navigator.api.nucleus.token import (
                TOKEN_FILE as navigator_token_file,
            )

        except ImportError:
            return None

        if service != KEYRING_NAME and username != auth_domain:
            return None
        else:
            try:
                os.remove(navigator_token_file)
            except FileNotFoundError:
                return None


class AnacondaKeyring(KeyringBackend):
    name = "token AnacondaKeyring" # Pinning name explicitly instead of relying on module.submodule automatic naming convention.
    keyring_path = Path("~/.anaconda/keyring").expanduser()

    @classproperty
    def priority(cls) -> float:
        config = AnacondaAuthConfig()
        if config.preferred_token_storage == "system":
            return 0.2
        elif config.preferred_token_storage == "anaconda-keyring":
            return 11.0
        else:
            raise ValueError(
                f"token_storage: {config.preferred_token_storage} is not supported."
            )

    @classproperty
    def viable(cls) -> bool:
        try:
            cls.keyring_path.parent.mkdir(exist_ok=True, parents=True)
            with cls.keyring_path.open("a") as f:
                writable = f.writable()
            return writable
        except OSError:
            return False

    def _read(self) -> LocalKeyringData:
        if not self.keyring_path.exists():
            return {}

        try:
            with self.keyring_path.open("r") as fp:
                data = json.load(fp)
            return data
        except json.JSONDecodeError:
            return {}

    def _save(self, data: LocalKeyringData) -> None:
        self.keyring_path.parent.mkdir(exist_ok=True, parents=True)

        with self.keyring_path.open("w") as fp:
            json.dump(data, fp)

    def set_password(self, service: str, username: str, password: str) -> None:
        data = self._read()

        if service not in data:
            data[service] = {}

        data[service][username] = password

        self._save(data)

    def get_password(self, service: str, username: str) -> Union[str, None]:
        data = self._read()
        return data.get(service, {}).get(username, None)

    def delete_password(self, service: str, username: str) -> None:
        data = self._read()
        try:
            data.get(service, {}).pop(username)
            self._save(data)
        except KeyError:
            raise PasswordDeleteError


class RepoToken(BaseModel):
    token: TokenString
    org_name: Union[OrgName, None] = None


# A mapping of modern domain to a list of legacy domains. If a token is searched
# for at the modern domain and not found, we will search for any of the legacy domains
# and, if found, migrate the keyring storage from that domain to the new one.
MIGRATIONS: Dict[str, List[str]] = {
    "anaconda.com": ["id.anaconda.cloud", "anaconda.cloud"]
}
TOKEN_INFO_VERSION = 2


class TokenInfo(BaseModel):
    domain: str = Field(default_factory=lambda: AnacondaAuthConfig().domain)
    api_key: Union[str, None] = None
    username: Union[str, None] = None
    repo_tokens: List[RepoToken] = []
    version: Optional[int] = TOKEN_INFO_VERSION

    @classmethod
    def _decode(cls, keyring_data: str) -> dict:
        decoded_bytes = base64.b64decode(keyring_data)
        decoded_dict = json.loads(decoded_bytes)
        return decoded_dict

    @classmethod
    def _migrate(
        cls, keyring_data: str, from_domain: str, to_domain: str
    ) -> "TokenInfo":
        """Migrate the domain and save token under new domain."""
        decoded_dict = cls._decode(keyring_data)
        decoded_dict["domain"] = to_domain
        decoded_dict["version"] = TOKEN_INFO_VERSION
        token_info = TokenInfo(**decoded_dict)
        token_info.save()
        keyring.delete_password(KEYRING_NAME, from_domain)
        logger.debug(
            f"🔓 Token has been migrated from legacy domain '{from_domain}' to '{to_domain}' 🎉"
        )
        return token_info

    @classmethod
    def load(cls, domain: Optional[str] = None, *, create: bool = False) -> "TokenInfo":
        """Load the token information from the system keyring.

        Args:
            domain: The domain for which to load the token information. If
                not provided, defaults to the configuration domain.
            create: If True, create a new TokenInfo object if not found.

        Returns:
            The token information.

        """
        domain = domain or AnacondaAuthConfig().domain

        keyring_data = keyring.get_password(KEYRING_NAME, domain)
        if keyring_data is not None:
            logger.debug("🔓 Token has been successfully retrieved from keyring 🎉")
            decoded_dict = cls._decode(keyring_data)
            return TokenInfo(**decoded_dict)

        # Try again to see if there is a legacy token on disk
        legacy_domains = MIGRATIONS.get(domain, [])
        for legacy_domain in legacy_domains:
            existing_keyring_data = keyring.get_password(KEYRING_NAME, legacy_domain)
            if existing_keyring_data is not None:
                return cls._migrate(
                    existing_keyring_data, from_domain=legacy_domain, to_domain=domain
                )

        if create:
            logger.debug("🔓 Token has been successfully created 🎉")
            return TokenInfo(domain=domain)

        raise TokenNotFoundError

    def save(self) -> None:
        """Write the token information to the system keyring."""
        payload = self.model_dump_json(exclude_none=True)
        encoded = _as_base64_string(payload)
        keyring.set_password(KEYRING_NAME, self.domain, encoded)
        logger.debug("🔒 Token has been safely stored in system keychain 🎉")

    def delete(self) -> None:
        """Delete the token information from the system keyring."""
        try:
            keyring.delete_password(KEYRING_NAME, self.domain)
            if NavigatorFallback.viable:
                NavigatorFallback().delete_password(KEYRING_NAME, self.domain)
        except PasswordDeleteError:
            raise TokenNotFoundError

    @property
    def expired(self) -> bool:
        if self.api_key is None:
            return True

        decoded = jwt.decode(
            self.api_key, algorithms=["RS256"], options={"verify_signature": False}
        )
        expiry = dt.datetime.fromtimestamp(decoded["exp"]).replace(
            tzinfo=dt.timezone.utc
        )
        return expiry < dt.datetime.now(tz=dt.timezone.utc)

    def get_access_token(self) -> str:
        """Get the access token, ensuring login and refresh if necessary."""
        if self.api_key is None:
            try:
                new_token_info = TokenInfo.load(self.domain)
            except TokenNotFoundError:
                message = "No token found, please login with `anaconda login`"
                raise TokenNotFoundError(message)

            # Store the new token information for later retrieval
            self.username = new_token_info.username
            self.api_key = new_token_info.api_key

        assert self.api_key is not None

        if self.expired:
            raise TokenExpiredError(
                "Your login token as expired. Please login again using\n"
                "  anaconda login --force"
            )

        return self.api_key

    def set_repo_token(self, org_name: OrgName, token: TokenString) -> None:
        """Set the repo token for a specific organization.

        Args:
            org_name: The organization name for which to search for a token.
            token: The token value.

        """

        try:
            self.get_repo_token(org_name=org_name)
        except TokenNotFoundError:
            # This is good, we don't need to do anything
            pass
        else:
            # We need to remove the existing token for this org first
            # TODO: We can drop this once we just use a dictionary instead
            self.repo_tokens[:] = [
                t for t in self.repo_tokens if t.org_name != org_name
            ]

        self.repo_tokens.append(RepoToken(org_name=org_name, token=token))

    def get_repo_token(self, org_name: OrgName) -> TokenString:
        """Return the installed repo token for a specific organization.

        Args:
            org_name: The organization name for which to search for a token.

        Returns:
            The repo access token associated with the requested organization.

        Raises:
            TokenNotFoundError: if no token is found in the keyring for that organization.

        """
        for token in self.repo_tokens:
            if token.org_name == org_name:
                return token.token
        raise TokenNotFoundError(f"Could not find repo token for org {org_name}")

    def delete_repo_token(self, org_name: OrgName) -> None:
        """Delete the repo token for a specific organization.

        Args:
            org_name: The organization name for which to delete a token.

        """
        # TODO: Confirm whether we should raise an exception or not if token doesn't exist for specified organization

        # We need to remove the existing token for this org first
        # TODO: We can drop this once we just use a dictionary instead
        self.repo_tokens[:] = [t for t in self.repo_tokens if t.org_name != org_name]



from datetime import datetime
from uuid import UUID
from uuid import uuid4

import jwt
import pytest
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker

from .conftest import CLIInvoker

pytest.importorskip("conda")

# ruff: noqa: E402
from anaconda_auth._conda.repo_config import REPO_URL
from anaconda_auth.repo import OrganizationData
from anaconda_auth.repo import RepoAPIClient
from anaconda_auth.repo import TokenCreateResponse
from anaconda_auth.repo import TokenInfoResponse
from anaconda_auth.token import TokenInfo
from anaconda_auth.token import TokenNotFoundError


@pytest.fixture()
def org_name() -> str:
    return "test-org-name"


@pytest.fixture()
def business_org_id() -> UUID:
    return uuid4()


@pytest.fixture()
def valid_api_key():
    token_info = TokenInfo.load(create=True)
    # The important part is that the key is not expired. If this still exists in
    # 2099, Trump hasn't blown up the world and it has far exceeded my expectations.
    token_info.api_key = jwt.encode(
        {"exp": datetime(2099, 1, 1).toordinal()}, key="secret", algorithm="HS256"
    )
    token_info.save()
    return token_info


@pytest.fixture(params=[True, False])
def no_tokens_installed(
    request, mocker: MockerFixture, valid_api_key: TokenInfo
) -> None:
    if request.param:
        # Remove the API key
        valid_api_key.delete()
    else:
        # Models the situation where we have a valid API key but it has no attached repo tokens.
        pass

    # No legacy tokens either
    mocker.patch(
        "anaconda_auth._conda.repo_config.read_binstar_tokens",
        return_value={},
    )


@pytest.fixture()
def token_is_installed(org_name: str, valid_api_key: TokenInfo) -> TokenInfo:
    valid_api_key.set_repo_token(org_name=org_name, token="test-token")
    valid_api_key.save()
    return valid_api_key


@pytest.fixture(autouse=True)
def mock_do_auth_flow(mocker: MockerFixture) -> None:
    mocker.patch(
        "anaconda_auth.repo._do_auth_flow",
        return_value="test-access-token",
    )


@pytest.fixture()
def token_does_not_exist_in_service(
    requests_mock: RequestMocker, org_name: str
) -> None:
    requests_mock.get(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        status_code=404,
    )


@pytest.fixture()
def token_exists_in_service(
    requests_mock: RequestMocker, org_name: str
) -> TokenInfoResponse:
    token_info = TokenInfoResponse(
        id=uuid4(), expires_at=datetime(year=2025, month=1, day=1)
    )
    requests_mock.get(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        json=token_info.model_dump(mode="json"),
    )
    return token_info


@pytest.fixture()
def token_created_in_service(
    requests_mock: RequestMocker, org_name: str
) -> TokenCreateResponse:
    test_token = "test-token"
    payload = {"token": test_token, "expires_at": "2025-01-01T00:00:00"}
    requests_mock.put(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        json=payload,
    )
    return TokenCreateResponse(**payload)


@pytest.fixture()
def user_has_one_org(
    requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> TokenCreateResponse:
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[
            {
                "id": str(business_org_id),
                "name": org_name,
                "title": "My Cool Organization",
            }
        ],
    )
    return [
        OrganizationData(
            id=business_org_id, name=org_name, title="My Cool Organization"
        )
    ]


@pytest.fixture()
def user_has_multiple_orgs(
    requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> TokenCreateResponse:
    first_id = uuid4()
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[
            {
                "id": str(first_id),
                "name": "first-org",
                "title": "My First Organization",
            },
            {
                "id": str(business_org_id),
                "name": org_name,
                "title": "My Business Organization",
            },
        ],
    )
    return [
        OrganizationData(id=first_id, name="first-org", title="My First Organizatoin"),
        OrganizationData(
            id=business_org_id, name=org_name, title="My Business Organization"
        ),
    ]


@pytest.fixture()
def user_has_no_orgs(
    requests_mock: RequestMocker, user_has_no_subscriptions: None
) -> list[OrganizationData]:
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[],
    )
    return []


@pytest.fixture(
    autouse=True, params=["security_subscription", "commercial_subscription"]
)
def user_has_business_subscription(
    request, requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> None:
    requests_mock.get(
        "https://anaconda.com/api/account",
        json={
            "subscriptions": [
                {
                    "org_id": str(business_org_id),
                    "product_code": request.param,
                }
            ]
        },
    )


@pytest.fixture()
def user_has_starter_subscription(
    request, requests_mock: RequestMocker, business_org_id: UUID
) -> None:
    requests_mock.get(
        "https://anaconda.com/api/account",
        json={
            "subscriptions": [
                {
                    "org_id": str(business_org_id),
                    "product_code": "starter_subscription",
                }
            ]
        },
    )


@pytest.fixture()
def user_has_no_subscriptions(requests_mock: RequestMocker) -> None:
    requests_mock.get("https://anaconda.com/api/account", json={})


@pytest.fixture(autouse=True)
def repodata_json_available_with_token(
    requests_mock: RequestMocker, token_created_in_service: TokenCreateResponse
) -> None:
    requests_mock.head(
        f"https://repo.anaconda.cloud/t/{token_created_in_service.token}/repo/main/noarch/repodata.json",
        status_code=200,
    )


def test_token_list_no_tokens(
    invoke_cli: CLIInvoker, no_tokens_installed: None
) -> None:
    result = invoke_cli(["token", "list"])

    assert result.exit_code == 1
    assert (
        "No repo tokens are installed. Run anaconda token install." in result.stdout
    ), result.stdout
    assert "Aborted." in result.stdout


def test_token_list_has_tokens(mocker: MockerFixture, invoke_cli: CLIInvoker) -> None:
    test_repo_token = "test-repo-token"
    mock = mocker.patch(
        "anaconda_auth._conda.repo_config.read_binstar_tokens",
        return_value={REPO_URL: test_repo_token},
    )
    result = invoke_cli(["token", "list"])

    mock.assert_called_once()

    assert result.exit_code == 0
    assert "Anaconda Repository Tokens" in result.stdout
    assert REPO_URL in result.stdout
    assert test_repo_token in result.stdout


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_does_not_exist_yet(
    option_flag: str,
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(
        ["token", "install", option_flag, org_name],
        input="y\n",
    )
    assert result.exit_code == 0

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_exists_already_accept(
    option_flag: str,
    org_name: str,
    token_exists_in_service: None,
    token_created_in_service: TokenCreateResponse,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install", option_flag, org_name], input="y\ny\n")
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_exists_already_decline(
    option_flag: str,
    org_name: str,
    valid_api_key: TokenInfo,
    token_exists_in_service: None,
    token_created_in_service: str,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install", option_flag, org_name], input="n")
    assert result.exit_code == 1, result.stdout

    token_info = TokenInfo.load()
    with pytest.raises(TokenNotFoundError):
        _ = token_info.get_repo_token(org_name=org_name)


def test_token_install_no_available_org(
    user_has_no_orgs: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install"])
    assert result.exit_code == 1, result.stdout
    assert "No organizations found." in result.stdout, result.stdout
    assert "Aborted." in result.stdout


def test_token_install_select_first_if_only_org(
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    user_has_one_org: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install"])
    assert result.exit_code == 0, result.stdout
    assert (
        f"Only one organization found, automatically selecting: {org_name}"
        in result.stdout
    )

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


def test_token_install_select_second_of_multiple_orgs(
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    user_has_multiple_orgs: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    # TODO: This uses the "j" key binding. I can't figure out how to send the right
    #       escape code for down arrow.
    result = invoke_cli(["token", "install"], input="j\n")
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_uninstall(
    option_flag: str,
    org_name: str,
    token_is_installed: TokenInfo,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "uninstall", option_flag, org_name])
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    with pytest.raises(TokenNotFoundError):
        _ = token_info.get_repo_token(org_name=org_name)


def test_get_repo_token_info_no_token(
    org_name: str, token_does_not_exist_in_service: None
) -> None:
    client = RepoAPIClient()
    token_info = client._get_repo_token_info(org_name=org_name)
    assert token_info is None


def test_get_repo_token_info_has_token(
    org_name: str,
    token_exists_in_service: TokenInfoResponse,
) -> None:
    client = RepoAPIClient()
    token_info = client._get_repo_token_info(org_name=org_name)
    assert token_info == token_exists_in_service


def test_create_repo_token_info_has_token(
    org_name: str,
    token_created_in_service: TokenCreateResponse,
) -> None:
    client = RepoAPIClient()
    token_info = client._create_repo_token(org_name=org_name)
    assert token_info == token_created_in_service


def test_get_organizations_for_user(user_has_one_org: list[OrganizationData]) -> None:
    client = RepoAPIClient()
    organizations = client.get_organizations_for_user()
    assert organizations == user_has_one_org


def test_get_business_organizations_for_user(
    org_name: str,
    business_org_id: UUID,
    user_has_multiple_orgs: list[OrganizationData],
    user_has_business_subscription: None,
) -> None:
    client = RepoAPIClient()
    organizations = client.get_business_organizations_for_user()
    assert organizations == [
        OrganizationData(
            id=business_org_id, name=org_name, title="My Business Organization"
        )
    ]


def test_get_business_organizations_for_user_only_starter(
    org_name: str,
    business_org_id: UUID,
    user_has_multiple_orgs: list[OrganizationData],
    user_has_starter_subscription: None,
) -> None:
    client = RepoAPIClient()
    organizations = client.get_business_organizations_for_user()
    assert organizations == []


def test_issue_new_token_prints_success_message_via_client(
    org_name: str, mocker: MockerFixture, capsys: pytest.CaptureFixture
) -> None:
    client = RepoAPIClient()
    mocker.patch.object(client, "_get_repo_token_info", return_value=None)

    mock_response = mocker.MagicMock()
    mock_response.expires_at = datetime(2025, 12, 31)
    mocker.patch.object(client, "_create_repo_token", return_value=mock_response)
    client.issue_new_token(org_name=org_name)
    res = capsys.readouterr()
    expected_msg = "Your conda token has been installed and expires 2025-12-31 00:00:00. To view your token(s), you can use anaconda token list\n"

    assert expected_msg in res.out


def test_issue_new_token_prints_success_message_via_cli(
    org_name: str,
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture,
    token_exists_in_service,
    token_created_in_service,
    invoke_cli,
) -> None:
    result = invoke_cli(["token", "install", "--org", org_name], input="y\nn\n")

    expected_msg = "Your conda token has been installed and expires 2025-01-01 00:00:00. To view your token(s), you can use anaconda token list\n"
    assert result.exit_code == 0, result.stdout
    assert expected_msg in result.stdout



from __future__ import annotations

from pathlib import Path
from typing import Any

from ruamel.yaml import YAML
from ruamel.yaml import YAMLError

from anaconda_cli_base import console

yaml = YAML()


class CondaRCError(Exception):
    pass


class CondaRC:
    def __init__(self, condarc_path: Path | None = None):
        """
        Initializes the CondaRC object by attempting to open and load the contents
        of the condarc file found in the user's home directory.
        """
        self.condarc_path = condarc_path or Path("~/.condarc").expanduser()
        self._loaded_yaml: dict[str, Any] = {}
        self.load()

    @property
    def _backup_condarc_path(self) -> Path:
        return self.condarc_path.with_suffix(".bak")

    def load(self, path: Path | None = None) -> None:
        path = path or self.condarc_path
        try:
            path.touch()
            with path.open("r") as fp:
                contents = fp.read()
        except OSError as exc:
            raise CondaRCError(f"Could not open condarc file: {exc}")

        try:
            self._loaded_yaml = yaml.load(contents) or {}
        except YAMLError as exc:
            raise CondaRCError(f"Could not parse condarc: {exc}")

    def update_channel_settings(
        self, channel: str, auth_type: str, username: str | None = None
    ) -> None:
        """
        Update the condarc file's "channel_settings" section
        """
        if username is None:
            updated_settings = {"channel": channel, "auth": auth_type}
        else:
            updated_settings = {
                "channel": channel,
                "auth": auth_type,
                "username": username,
            }

        channel_settings = self._loaded_yaml.get("channel_settings", []) or []

        # Filter out the existing channel's entry if it's there
        filter_settings = [
            settings
            for settings in channel_settings
            if settings.get("channel") != channel
        ]

        # Add the updated settings map
        filter_settings.append(updated_settings)

        self._loaded_yaml["channel_settings"] = filter_settings

    def restore(self) -> None:
        console.print(f"Restoring config from {self._backup_condarc_path}")
        self.load(self._backup_condarc_path)
        self.save()

    def backup(self) -> None:
        console.print(f"Backing up config to {self._backup_condarc_path}")
        self.save(self._backup_condarc_path)

    def save(self, path: Path | None = None) -> None:
        """Save the condarc file"""
        path = path or self.condarc_path
        try:
            with path.open("w") as fp:
                yaml.dump(self._loaded_yaml, fp)
        except OSError as exc:
            raise CondaRCError(f"Could not save file: {exc}")



from __future__ import annotations

import os
from typing import Any
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from anaconda_auth import __version__
from anaconda_auth import login
from anaconda_auth.actions import get_api_key
from anaconda_auth.actions import is_logged_in
from anaconda_auth.client import BaseClient
from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.token import TOKEN_INFO_VERSION
from anaconda_auth.token import TokenInfo

from .conftest import MockedRequest

HERE = os.path.dirname(__file__)


def test_login_to_api_key(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.actions.get_api_key", return_value="api-key")
    mocker.patch("anaconda_auth.actions._do_auth_flow")

    login()

    config = AnacondaAuthConfig()
    keyring_token = TokenInfo.load(config.domain)

    assert keyring_token.model_dump() == {
        "domain": config.domain,
        "username": None,
        "repo_tokens": [],
        "api_key": "api-key",
        "version": TOKEN_INFO_VERSION,
    }


def test_login_ssl_verify(mocker: MockerFixture, api_key: str) -> None:
    mocker.patch("anaconda_auth.actions.get_api_key", return_value=api_key)
    do_auth_flow = mocker.patch("anaconda_auth.actions._do_auth_flow")

    login(ssl_verify=True)
    assert do_auth_flow.call_args_list[-1].kwargs["config"].ssl_verify


def test_login_no_ssl_verify(mocker: MockerFixture, api_key: str) -> None:
    mocker.patch("anaconda_auth.actions.get_api_key", return_value=api_key)
    do_auth_flow = mocker.patch("anaconda_auth.actions._do_auth_flow")

    login(ssl_verify=False)
    assert not do_auth_flow.call_args_list[-1].kwargs["config"].ssl_verify


@pytest.mark.integration
def test_get_auth_info(integration_test_client: BaseClient, is_not_none: Any) -> None:
    response = integration_test_client.get("/api/account")
    assert response.status_code == 200
    assert response.json() == {
        "user": is_not_none,
        "profile": is_not_none,
        "subscriptions": is_not_none,
    }


@pytest.fixture
def mocked_do_login(mocker: MockerFixture) -> MagicMock:
    def _mocked_login(config: AnacondaAuthConfig, basic: bool) -> None:
        TokenInfo(domain=config.domain, api_key="from-login").save()

    mocker.patch("anaconda_auth.actions._do_login", _mocked_login)
    from anaconda_auth import actions

    login_spy = mocker.spy(actions, "_do_login")
    return login_spy


def test_login_no_existing_token(mocked_do_login: MagicMock) -> None:
    config = AnacondaAuthConfig()
    login(config=config)

    assert TokenInfo.load(config.domain).api_key == "from-login"
    mocked_do_login.assert_called_once()


def test_login_has_valid_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    config = AnacondaAuthConfig()

    mocker.patch("anaconda_auth.token.TokenInfo.expired", False)
    TokenInfo(domain=config.domain, api_key="pre-existing").save()

    login(config=config)
    mocked_do_login.assert_not_called()

    assert TokenInfo.load(config.domain).api_key == "pre-existing"


def test_force_login_with_valid_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    config = AnacondaAuthConfig()

    mocker.patch("anaconda_auth.token.TokenInfo.expired", False)
    TokenInfo(domain=config.domain, api_key="pre-existing").save()

    login(config=config, force=True)
    mocked_do_login.assert_called_once()

    assert TokenInfo.load(config.domain).api_key == "from-login"


def test_login_has_expired_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    config = AnacondaAuthConfig()

    mocker.patch("anaconda_auth.token.TokenInfo.expired", True)
    TokenInfo(domain=config.domain, api_key="pre-existing-expired").save()

    login(config=config)
    mocked_do_login.assert_called_once()

    assert TokenInfo.load(config.domain).api_key == "from-login"


@pytest.fixture()
def mocked_request(mocker: MockerFixture) -> MockedRequest:
    """A mocked post request returning an API key."""

    # This could be generalized further, but it may not be worth the effort
    # For now, this mimics a fixed POST request with fixed mocked return data

    mocked_request = MockedRequest(
        response_status_code=201, response_data={"api_key": "some-jwt"}
    )
    mocker.patch("requests.post", mocked_request)
    return mocked_request


@pytest.mark.usefixtures("without_aau_token")
def test_get_api_key(mocked_request: MockedRequest) -> None:
    """When we get an API key, we assign appropriate generic scopes and tags."""

    key = get_api_key("some_access_token")
    assert key == "some-jwt"

    headers = mocked_request.called_with_kwargs["headers"]
    assert headers["Authorization"].startswith("Bearer")
    assert "X-AAU-CLIENT" not in headers

    data = mocked_request.called_with_kwargs["json"]
    assert data == {
        "scopes": ["cloud:read", "cloud:write", "repo:read"],
        "tags": [f"anaconda-auth/v{__version__}"],
    }

@pytest.mark.usefixtures("without_aau_token")
def test_get_api_key_with_custom_config(mocked_request: MockedRequest) -> None:
    """When we get an API key, we assign appropriate generic scopes and tags."""

    config = AnacondaAuthConfig(auth_domain_override="auth.example.domain")

    key = get_api_key("some_access_token", config=config)
    assert key == "some-jwt"
    assert mocked_request.called_with_args == ("https://auth.example.domain/api/auth/api-keys",)

    headers = mocked_request.called_with_kwargs["headers"]
    assert headers["Authorization"].startswith("Bearer")
    assert "X-AAU-CLIENT" not in headers

    data = mocked_request.called_with_kwargs["json"]
    assert data == {
        "scopes": ["cloud:read", "cloud:write", "repo:read"],
        "tags": [f"anaconda-auth/v{__version__}"],
    }

@pytest.mark.usefixtures("with_aau_token")
def test_get_api_key_with_aau_token(mocked_request: MockedRequest) -> None:
    """When we get an API key, we assign appropriate generic scopes and tags."""

    key = get_api_key("some_access_token")
    assert key == "some-jwt"

    headers = mocked_request.called_with_kwargs["headers"]
    assert headers["Authorization"].startswith("Bearer")
    assert headers["X-AAU-CLIENT"] == "anon-token"


@pytest.mark.usefixtures("save_api_key_to_token")
def test_is_logged_in() -> None:
    assert is_logged_in()



import importlib

import pytest

# A list of all public imports, relative to the package name
PUBLIC_IMPORTS = [
    "__version__",
    "login",
    "logout",
    "client_factory",
    "actions._do_auth_flow",
    "actions.get_api_key",
    "actions.is_logged_in",
    "actions.login",
    "actions.logout",
    "actions.make_auth_code_request_url",
    "actions.refresh_access_token",
    "actions.request_access_token",
    "client.BaseClient",
    "client.BearerAuth",
    "client.client_factory",
    "client.login_required",
    # TODO: figure out what to do with this import
    # "config.AnacondaCloudConfig",
    "exceptions.AuthenticationError",
    "exceptions.InvalidTokenError",
    "exceptions.TokenNotFoundError",
    "exceptions.LoginRequiredError",
    "exceptions.TokenExpiredError",
    "handlers.capture_auth_code",
    "handlers.shutdown_all_servers",
    "token.TokenInfo",
]


@pytest.fixture(autouse=True)
def reset_imports():
    yield
    importlib.invalidate_caches()


@pytest.mark.parametrize(
    "rel_attr_path",
    PUBLIC_IMPORTS,
)
def test_import_aliases(rel_attr_path):
    """Given a relative nested import, ensure it's the same for both anaconda_auth and anaconda_cloud_auth."""
    sub_mod_path, _, attr_name = rel_attr_path.rpartition(".")

    mod_path = "anaconda_cloud_auth" + (f".{sub_mod_path}" if sub_mod_path else "")
    mod = importlib.import_module(mod_path)
    val_1 = getattr(mod, attr_name)

    mod_path = "anaconda_auth" + (f".{sub_mod_path}" if sub_mod_path else "")
    mod = importlib.import_module(mod_path)
    val_2 = getattr(mod, attr_name)

    assert val_1 is val_2


@pytest.mark.parametrize(
    "rel_attr_path",
    PUBLIC_IMPORTS,
)
def test_deprecation_warning(rel_attr_path):
    sub_mod_path, _, attr_name = rel_attr_path.rpartition(".")

    mod_path = "anaconda_cloud_auth" + (f".{sub_mod_path}" if sub_mod_path else "")

    mod = importlib.import_module(mod_path)
    with pytest.deprecated_call():
        importlib.reload(mod)



"""
CLI for conda-token.
"""

from __future__ import annotations

import os
import sys
from argparse import ArgumentParser
from argparse import Namespace

from anaconda_auth import __version__
from anaconda_auth._conda import repo_config


def token_list(args: Namespace) -> int:
    """Default function for list subparser."""
    tokens = {
        k: v for k, v in repo_config.token_list().items() if k == repo_config.REPO_URL
    }
    if not tokens:
        print(
            f"No tokens have been configured for {repo_config.REPO_URL}",
            file=sys.stderr,
        )
        return 1

    for url, token in tokens.items():
        print(url, token)

    return 0


def token_set(args: Namespace) -> int:
    try:
        repo_config.validate_token(args.token, no_ssl_verify=args.no_ssl_verify)
    except repo_config.CondaTokenError as e:
        print(e, file=sys.stderr)
        return 1

    repo_config.token_set(
        args.token,
        args.system,
        args.env,
        args.file,
        args.include_archive_channels,
        args.no_ssl_verify,
        args.enable_signature_verification,
        force=args.force_config_condarc,
    )
    print("Success! Your token was validated and Conda has been configured.")
    return 0


def token_remove(args: Namespace) -> int:
    repo_config.token_remove()
    return 0


def condarc_path_args(parser: ArgumentParser) -> None:
    """Add condarc path arguments."""
    config_file_location_group = parser.add_argument_group(
        "Config File Location Selection",
        f"Without one of these flags, the user config file at '{repo_config.escaped_user_rc_path}' is used.",
    )
    location = config_file_location_group.add_mutually_exclusive_group()
    location.add_argument(
        "--system",
        action="store_true",
        help=f"Write to the system .condarc file at '{repo_config.escaped_user_rc_path}'.",
    )
    location.add_argument(
        "--env",
        action="store_true",
        help=(
            f"Write to the active conda environment .condarc file ({os.getenv('CONDA_PREFIX', '<no active environment>').replace('%', '%%')}). "
            f"If no environment is active, write to the user config file ({repo_config.escaped_user_rc_path})."
        ),
    )
    location.add_argument("--file", action="store", help="Write to the given file.")


def cli(argv: list[str] | None = None) -> int:
    parser = ArgumentParser(
        "conda-token",
        usage="conda token",
        description="Configure token access for Anaconda Commercial Edition",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="Show the conda-token version number and exit.",
        version=f"conda-token {__version__}",
    )

    subparser = parser.add_subparsers(help="Token commands")

    subparser_list = subparser.add_parser("list", help="List token if configured.")
    subparser_list.set_defaults(func=token_list)

    subparser_remove = subparser.add_parser(
        "remove", help="Remove token and revert default_channels."
    )
    condarc_path_args(subparser_remove)
    subparser_remove.set_defaults(func=token_remove)

    subparser_set = subparser.add_parser(
        "set", help="Set your token and configure default_channels."
    )
    subparser_set.add_argument("token", help="Your token.")
    subparser_set.add_argument(
        "--include-archive-channels",
        choices=repo_config.ARCHIVE_CHANNELS,
        help="Add archived channels to default_channels. "
        "\nAvailable channels are mro, mro-archive, free, and pro.",
        nargs="+",
        default=None,
        metavar="CHANNEL_NAME",
    )
    subparser_set.add_argument(
        "--no-ssl-verify",
        help="Force disabling SSL verification. "
        "By default conda-token will respect your "
        "ssl_verify flag in your Conda settings. "
        "Using this flag will add ssl_verify: false "
        "to your .condarc file. If not already set.",
        action="store_true",
    )
    subparser_set.add_argument(
        "--enable-signature-verification",
        action="store_true",
        help="Enable Conda package signature verification. Default is off.",
    )
    subparser_set.add_argument(
        "--force-config-condarc",
        action="store_true",
        help="Force configuration of .condarc",
    )
    condarc_path_args(subparser_set)
    subparser_set.set_defaults(func=token_set)

    if len(sys.argv) == 1:
        sys.argv.append("--help")

    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    cli(sys.argv[1:])



import pytest
import urllib3.exceptions
from packaging.version import parse

from anaconda_auth._conda.conda_token import cli
from anaconda_auth._conda.repo_config import CONDA_VERSION
from anaconda_auth._conda.repo_config import CondaVersionWarning


@pytest.mark.skip(reason="blocking release in CI but passing fine locally")
def test_token_set_no_verify_ssl(remove_token_no_repo_url_mock, secret_token, capsys):
    # real InsecureRequestWarning against real server
    with pytest.warns(urllib3.exceptions.InsecureRequestWarning):
        cli(["set", "--no-ssl-verify", secret_token, "--force-config-condarc"])


@pytest.mark.skip(reason="blocking release in CI but passing fine locally")
def test_token_set_no_verify_ssl_mock_server(
    remove_token, secret_token, capsys, repo_url
):
    cli(["set", "--no-ssl-verify", "--force-config-condarc", secret_token])
    ret = cli(["list"])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out.splitlines()[-1] == repo_url + " " + secret_token


def test_token_list(remove_token, capsys, repo_url):
    ret = cli(["list"])
    captured = capsys.readouterr()
    assert ret == 1
    assert (
        captured.err.splitlines()[-1]
        == f"No tokens have been configured for {repo_url}"
    )


def test_token_set_invalid_channel(remove_token):
    with pytest.raises(SystemExit):
        cli(["set", "secret", "--include-archive-channels", "nope"])


@pytest.mark.skip(reason="blocking release in CI but passing fine locally")
def test_token_set(remove_token, secret_token, capsys, repo_url):
    cli(["set", "--force-config-condarc", secret_token])

    ret = cli(["list"])
    assert ret == 0
    captured = capsys.readouterr()
    assert (
        "Success! Your token was validated and Conda has been configured."
        in captured.out
    )


def test_token_set_error(remove_token, capsys, repo_url):
    ret = cli(["set", "SECRET"])
    assert ret == 1
    captured = capsys.readouterr()
    # we will also capture a HTTP log for the mock server; check the last line.
    assert (
        captured.err.splitlines()[-1]
        == "The token could not be validated. Please check that you have typed it correctly."
    )

    ret = cli(["list"])
    assert ret == 1
    captured = capsys.readouterr()
    assert (
        captured.err.splitlines()[-1]
        == f"No tokens have been configured for {repo_url}"
    )


@pytest.mark.skipif(
    CONDA_VERSION >= parse("4.10.1"),
    reason="Signature verification will warn on old versions",
)
def test_token_set_with_signing_warn(remove_token, secret_token, capsys):
    with pytest.warns(CondaVersionWarning):
        ret = cli(["set", "--enable-signature-verification", secret_token])
        assert ret == 0



import sys
import warnings
from textwrap import dedent
from typing import Optional

import typer
from requests.exceptions import HTTPError
from requests.exceptions import JSONDecodeError
from rich.prompt import Confirm
from rich.syntax import Syntax

from anaconda_auth import __version__
from anaconda_auth.actions import login
from anaconda_auth.actions import logout
from anaconda_auth.client import BaseClient
from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.exceptions import TokenExpiredError
from anaconda_auth.token import TokenInfo
from anaconda_auth.token import TokenNotFoundError
from anaconda_cli_base.config import anaconda_config_path
from anaconda_cli_base.console import console
from anaconda_cli_base.exceptions import register_error_handler


def _continue_with_login() -> int:
    if sys.stdout.isatty():
        do_login = Confirm.ask("Continue with interactive login?", choices=["y", "n"])
        if do_login:
            login()
            return -1
        else:
            console.print(
                dedent("""
                To configure your credentials you can run
                  [green]anaconda login --at anaconda.com[/green]

                or set your API key using the [green]ANACONDA_AUTH_API_KEY[/green] env var

                or set
                """)
            )
            console.print(
                Syntax(
                    dedent(
                        """\
                        [plugin.auth]
                        api_key = "<api-key>"
                        """
                    ),
                    "toml",
                    background_color=None,
                )
            )
            console.print(f"in {anaconda_config_path()}")
    return 1


def _login_required_message(error_classifier: str) -> None:
    console.print(
        f"[bold][red]{error_classifier}[/red][/bold]: Login is required to complete this action."
    )


@register_error_handler(TokenNotFoundError)
def login_required(e: Exception) -> int:
    _login_required_message(e.__class__.__name__)
    return _continue_with_login()


@register_error_handler(TokenExpiredError)
def token_expired(e: Exception) -> int:
    console.print(
        f"[bold][red]{e.__class__.__name__}[/red][/bold]: Your login token has expired"
    )

    return _continue_with_login()


@register_error_handler(HTTPError)
def http_error(e: HTTPError) -> int:
    try:
        error_code = e.response.json().get("error", {}).get("code", "")
    except JSONDecodeError:
        error_code = ""

    if error_code == "auth_required":
        if "Authorization" in e.request.headers:
            console.print(
                "[bold][red]InvalidAuthentication:[/red][/bold] Your provided API Key or login token is invalid"
            )
        else:
            _login_required_message("AuthenticationMissingError")
        return _continue_with_login()
    else:
        console.print(f"[bold][red]{e.__class__.__name__}:[/red][/bold] {e}")
        return 1


app = typer.Typer(name="auth", add_completion=False, help="anaconda.com auth commands")


@app.callback(invoke_without_command=True)
def main(
    version: bool = typer.Option(False, "-V", "--version"),
    name: Optional[str] = typer.Option(
        None,
        "-n",
        "--name",
        hidden=True,
    ),
    organization: Optional[str] = typer.Option(
        None,
        "-o",
        "--org",
        "--organization",
        hidden=True,
    ),
    strength: Optional[str] = typer.Option(
        None,
        "--strength",
        hidden=True,
    ),
    strong: Optional[bool] = typer.Option(
        None,
        "--strong",
        hidden=True,
    ),
    weak: Optional[bool] = typer.Option(
        None,
        "--strong",
        "-w",
        "--weak",
        hidden=True,
    ),
    url: Optional[str] = typer.Option(
        None,
        "--url",
        hidden=True,
    ),
    max_age: Optional[str] = typer.Option(
        None,
        "--max-age",
        hidden=True,
    ),
    scopes: Optional[str] = typer.Option(
        None,
        "-s",
        "--scopes",
        hidden=True,
    ),
    out: Optional[str] = typer.Option(
        None,
        "--out",
        hidden=True,
    ),
    list_scopes: Optional[bool] = typer.Option(
        None,
        "-x",
        "--list-scopes",
        hidden=True,
    ),
    list_tokens: Optional[bool] = typer.Option(
        None,
        "-l",
        "--list",
        hidden=True,
    ),
    remove: Optional[str] = typer.Option(
        None,
        "-r",
        "--remove",
        hidden=True,
    ),
    create: Optional[bool] = typer.Option(
        None,
        "-c",
        "--create",
        hidden=True,
    ),
    info: Optional[bool] = typer.Option(
        None,
        "-i",
        "--info",
        "--current-info",
        hidden=True,
    ),
) -> None:
    if version:
        console.print(
            f"anaconda-auth, version [cyan]{__version__}[/cyan]",
            style="bold green",
        )
        raise typer.Exit()

    has_options = any(
        value is not None
        for value in (
            name,
            organization,
            strength,
            strong,
            weak,
            url,
            max_age,
            scopes,
            out,
            list_scopes,
            list_tokens,
            remove,
            create,
            info,
        )
    )
    if has_options:
        # If any of the anaconda-client options are passed, try to delegate to
        # binstar_main if it exists. Otherwise, we just exit gracefully.

        try:
            from binstar_client.scripts.cli import main as binstar_main
        except (ImportError, ModuleNotFoundError):
            return

        console.print(
            "[yellow]DeprecationWarning[/yellow]: Please use [cyan]anaconda org auth[/cyan] instead for explicit management of anaconda.org auth tokens\n"
        )
        warnings.warn(
            "Please use `anaconda org auth` instead for explicit management of anaconda.org auth tokens",
            DeprecationWarning,
        )

        binstar_main(sys.argv[1:], allow_plugin_main=False)


@app.command("login")
def auth_login(force: bool = False, ssl_verify: bool = True) -> None:
    """Login"""
    try:
        auth_domain = AnacondaAuthConfig().domain
        expired = TokenInfo.load(domain=auth_domain).expired
        if expired:
            console.print("Your API key has expired, logging into anaconda.com")
            login(force=True, ssl_verify=ssl_verify)
            raise typer.Exit()
    except TokenNotFoundError:
        pass  # Proceed to login
    else:
        force = force or Confirm.ask(
            f"You are already logged into Anaconda ({auth_domain}). Would you like to force a new login?",
            default=False,
        )
        if not force:
            raise typer.Exit()

    login(force=force, ssl_verify=ssl_verify)


@app.command(name="whoami")
def auth_info() -> None:
    """Display information about the currently signed-in user"""
    client = BaseClient()
    response = client.get("/api/account")
    response.raise_for_status()
    console.print("Your anaconda.com info:")
    console.print_json(data=response.json(), indent=2, sort_keys=True)


@app.command(name="api-key")
def auth_key() -> None:
    """Display API Key for signed-in user"""
    config = AnacondaAuthConfig()
    if config.api_key:
        print(config.api_key)
        return

    token_info = TokenInfo.load(domain=config.domain)
    if not token_info.expired:
        print(token_info.api_key)
        return
    else:
        raise TokenExpiredError()


@app.command(name="logout")
def auth_logout() -> None:
    """Logout"""
    logout()



"""This module provides conda plugins.

Since conda is not a strict dependency of `anaconda-auth`, the modules here should
NOT be imported outside of this module. This package is declared as the plugin entrypoint
inside `pyproject.toml` such that it will be imported and, if conda is installed, the
plugin definitions will be imported. However, if conda is not installed, no plugins are
imported into the runtime.

As a result, we do not have to spread conditional imports everywhere, but we do need to
be careful that only the single top-level import of this subpackage is performed.

"""



"""
Minimal anaconda.cloud implementation for testing.
"""

import io
import threading
import wsgiref.simple_server
import wsgiref.util

MESSAGES = io.StringIO()


class QuietHandler(wsgiref.simple_server.WSGIRequestHandler):
    def log_message(self, format, *args):
        """Log an arbitrary message (don't confuse test's stdout/stderr capture)."""
        message = format % args
        MESSAGES.write(
            f"{self.address_string()} - - [{self.log_date_time_string()}] {message.translate(self._control_char_table)}\n"
        )


class App:
    def __call__(self, environ, start_response):
        request_path = wsgiref.util.request_uri(environ, include_query=False)
        # require token, but disallow the invalid token
        if "/t/" in request_path and "/t/SECRET/" not in request_path:
            status = "200 OK"
        else:
            status = "403 Forbidden"
        response_body = f"Request method: {environ['REQUEST_METHOD']}"
        response_headers = [
            ("Content-Type", "text/plain"),
            ("Content-Length", str(len(response_body))),
        ]
        start_response(status, response_headers)
        return [response_body.encode()]


def run_server():
    app = App()  # wsgiref.types added in 3.11
    server = wsgiref.simple_server.make_server(
        "127.0.0.1", 0, app, handler_class=QuietHandler
    )  # type: ignore
    address, port = server.socket.getsockname()
    t = threading.Thread(target=server.serve_forever, daemon=True)
    t.start()
    return f"http://{address}:{port}"


if __name__ == "__main__":
    run_server()
    import time

    time.sleep(300)



from anaconda_auth.client import *  # noqa: F403

# The following imports address a bug in Navigator, which imported from the wrong module by mistake
from anaconda_auth.config import AnacondaCloudConfig  # noqa: F401
from anaconda_auth.token import TokenInfo  # noqa: F401

# This one needs to stay to raise the deprecation warning
from anaconda_cloud_auth import warn  # noqa: F401

warn()



from __future__ import annotations

import os
import warnings
from uuid import uuid4

import pytest
from pytest import MonkeyPatch
from pytest_mock import MockerFixture
from requests import Request
from requests.exceptions import SSLError

from anaconda_auth.client import BaseClient
from anaconda_auth.client import client_factory
from anaconda_auth.token import TokenInfo

from .conftest import MockedRequest

HERE = os.path.dirname(__file__)


@pytest.mark.integration
@pytest.mark.usefixtures("disable_dot_env")
def test_login_required_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)

    client = BaseClient()

    request = Request("GET", "api/account")
    prepped = client.prepare_request(request)
    assert "Authorization" not in prepped.headers

    res = client.send(prepped)
    assert not res.ok
    assert "must login" in res.reason


@pytest.mark.integration
@pytest.mark.usefixtures("disable_dot_env")
def test_outdated_api_key(outdated_api_key: str) -> None:
    client = BaseClient(api_key=outdated_api_key)

    request = Request("GET", "api/account")
    prepped = client.prepare_request(request)
    assert prepped.headers.get("Authorization") == f"Bearer {outdated_api_key}"

    res = client.send(prepped)
    assert not res.ok
    assert "is invalid" in res.reason


@pytest.mark.integration
@pytest.mark.usefixtures("disable_dot_env")
def test_expired_token_ignored(
    outdated_token_info: TokenInfo, monkeypatch: MonkeyPatch
) -> None:
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)

    outdated_token_info.save()

    client = BaseClient(domain="mocked-domain")
    request = Request("GET", "api/account")
    prepped = client.prepare_request(request)
    assert "Authorization" not in prepped.headers


def test_client_factory_user_agent() -> None:
    client = client_factory("my-app/version")
    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("User-Agent") == "my-app/version"
    assert "Api-Version" not in response.request.headers


def test_client_factory_api_version() -> None:
    client = client_factory(user_agent="my-app/version", api_version="2023.01.01")
    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("User-Agent") == "my-app/version"
    assert response.request.headers.get("Api-Version") == "2023.01.01"


def test_client_subclass_api_version() -> None:
    class Client(BaseClient):
        _user_agent = "my-app/version"
        _api_version = "2023.01.01"

    client = Client()
    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("User-Agent") == "my-app/version"
    assert response.request.headers.get("Api-Version") == "2023.01.01"


@pytest.mark.parametrize(
    "attr_name, value, expected_base_uri",
    [
        ("domain", "anaconda.cloud", "https://anaconda.cloud"),
        ("domain", "dev.anaconda.cloud", "https://dev.anaconda.cloud"),
        ("base_uri", "https://anaconda.cloud", "https://anaconda.cloud"),
        ("base_uri", "https://dev.anaconda.cloud", "https://dev.anaconda.cloud"),
    ],
)
def test_client_base_uri(attr_name: str, value: str, expected_base_uri: str) -> None:
    client = BaseClient(**{attr_name: value})  # type: ignore
    assert client._base_uri == expected_base_uri


def test_client_base_uri_and_domain_raises_error() -> None:
    with pytest.raises(ValueError):
        BaseClient(domain="anaconda.cloud", base_uri="https://anaconda.cloud")


@pytest.fixture()
def mocked_request(mocker: MockerFixture) -> MockedRequest:
    """A mocked request, returning a custom response."""

    mocked_request = MockedRequest(
        response_status_code=200, response_headers={"Min-Api-Version": "2023.02.02"}
    )
    mocker.patch("requests.Session.request", mocked_request)
    return mocked_request


@pytest.mark.usefixtures("mocked_request")
@pytest.mark.parametrize(
    "api_version, warning_expected", [("2023.01.01", True), ("2023.03.01", False)]
)
def test_client_min_api_version_header(
    api_version: str, warning_expected: bool
) -> None:
    client = BaseClient(user_agent="client/0.1.0", api_version=api_version)
    with warnings.catch_warnings(record=True) as w:
        response = client.get("/api/something")

    assert response.status_code == 200
    assert response.headers.get("Min-Api-Version") == "2023.02.02"

    if warning_expected:
        assert len(w) == 1
        assert issubclass(w[0].category, DeprecationWarning)
        assert (
            "Client API version is 2023.01.01, minimum supported API version is 2023.02.02. "
            "You may need to update your client." == str(w[0].message)
        )
    else:
        assert len(w) == 0


@pytest.mark.usefixtures("disable_dot_env")
def test_anonymous_endpoint(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)

    # TODO: Use a mock for the request
    client = BaseClient()
    request = Request("GET", "api/projects/healthz")
    prepped = client.prepare_request(request)
    assert "Authorization" not in prepped.headers

    res = client.send(prepped)
    assert res


@pytest.mark.usefixtures("disable_dot_env")
def test_token_included(
    mocker: MockerFixture,
    monkeypatch: MonkeyPatch,
    outdated_token_info: TokenInfo,
) -> None:
    monkeypatch.setenv("ANACONDA_AUTH_DOMAIN", "mocked-domain")
    mocker.patch("anaconda_auth.token.TokenInfo.expired", False)
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)

    outdated_token_info.save()

    client = BaseClient()
    request = Request("GET", "api/catalogs/examples")
    prepped = client.prepare_request(request)
    assert prepped.headers["Authorization"] == f"Bearer {outdated_token_info.api_key}"


def test_api_key_env_variable_over_keyring(
    outdated_token_info: TokenInfo, monkeypatch: MonkeyPatch
) -> None:
    outdated_token_info.save()
    monkeypatch.setenv("ANACONDA_AUTH_API_KEY", "set-in-env")

    client = BaseClient()
    assert client.config.api_key == "set-in-env"

    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("Authorization") == "Bearer set-in-env"


def test_api_key_init_arg_over_variable(
    outdated_token_info: TokenInfo, monkeypatch: MonkeyPatch
) -> None:
    outdated_token_info.save()
    monkeypatch.setenv("ANACONDA_AUTH_API_KEY", "set-in-env")

    client = BaseClient(api_key="set-in-init")
    assert client.config.api_key == "set-in-init"

    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("Authorization") == "Bearer set-in-init"


def test_name_reverts_to_email(mocker: MockerFixture) -> None:
    account = {
        "user": {
            "id": "uuid",
            "email": "me@example.com",
            "first_name": None,
            "last_name": None,
        }
    }

    mocker.patch(
        "anaconda_auth.client.BaseClient.account",
        return_value=account,
        new_callable=mocker.PropertyMock,
    )
    client = BaseClient()

    assert client.email == "me@example.com"
    assert client.name == client.email


def test_first_and_last_name(mocker: MockerFixture) -> None:
    account = {
        "user": {
            "id": "uuid",
            "email": "me@example.com",
            "first_name": "Anaconda",
            "last_name": "User",
        }
    }

    mocker.patch(
        "anaconda_auth.client.BaseClient.account",
        return_value=account,
        new_callable=mocker.PropertyMock,
    )
    client = BaseClient()

    assert client.email == "me@example.com"
    assert client.name == "Anaconda User"


def test_gravatar_missing(mocker: MockerFixture) -> None:
    account = {
        "user": {
            "id": "uuid",
            "email": f"{uuid4()}@example.com",
            "first_name": "Anaconda",
            "last_name": "User",
        }
    }

    mocker.patch(
        "anaconda_auth.client.BaseClient.account",
        return_value=account,
        new_callable=mocker.PropertyMock,
    )
    client = BaseClient()

    assert client.avatar is None


def test_gravatar_found(mocker: MockerFixture) -> None:
    account = {
        "user": {
            "id": "uuid",
            "email": "test1@example.com",
            "first_name": "Anaconda",
            "last_name": "User",
        }
    }

    mocker.patch(
        "anaconda_auth.client.BaseClient.account",
        return_value=account,
        new_callable=mocker.PropertyMock,
    )
    client = BaseClient()
    assert client.avatar is not None


def test_extra_headers_dict() -> None:
    extra_headers = {"X-Extra": "stuff"}
    client = BaseClient(api_version="ver", extra_headers=extra_headers)

    res = client.get("api/something")
    assert res.request.headers["X-Extra"] == "stuff"
    assert res.request.headers["Api-Version"] == "ver"


def test_extra_headers_string() -> None:
    extra_headers = '{"X-Extra": "stuff"}'
    client = BaseClient(api_version="ver", extra_headers=extra_headers)

    res = client.get("api/something")
    assert res.request.headers["X-Extra"] == "stuff"
    assert res.request.headers["Api-Version"] == "ver"


def test_extra_headers_non_overwrite() -> None:
    extra_headers = {"X-Extra": "stuff", "Api-Version": "never overwrite"}
    client = BaseClient(api_version="ver", extra_headers=extra_headers)

    res = client.get("api/something")
    assert res.request.headers["X-Extra"] == "stuff"
    assert res.request.headers["Api-Version"] == "ver"


def test_extra_headers_bad_json() -> None:
    extra_headers = "nope"

    with pytest.raises(ValueError):
        _ = BaseClient(api_version="ver", extra_headers=extra_headers)


def test_extra_headers_env_var(monkeypatch: MonkeyPatch) -> None:
    extra_headers = '{"X-Extra": "from-env"}'
    monkeypatch.setenv("ANACONDA_AUTH_EXTRA_HEADERS", extra_headers)

    client = BaseClient(api_key="set-in-init")

    res = client.get("api/something")
    assert res.request.headers["X-Extra"] == "from-env"


@pytest.mark.integration
@pytest.mark.usefixtures("save_api_key_to_token")
def test_client_ssl_verify_true(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv(
        "REQUESTS_CA_BUNDLE", os.path.join(HERE, "resources", "mock-cert.pem")
    )

    client = BaseClient(ssl_verify=True)
    with pytest.raises(SSLError):
        client.get("api/account")


@pytest.mark.integration
@pytest.mark.usefixtures("save_api_key_to_token")
def test_login_ssl_verify_false(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv(
        "REQUESTS_CA_BUNDLE", os.path.join(HERE, "resources", "mock-cert.pem")
    )

    client = BaseClient(ssl_verify=False)
    res = client.get("api/account")
    assert res.ok


@pytest.mark.parametrize(
    "hash,hostname,expected_result",
    [
        (False, "test-hostname", "test-hostname"),
        (True, "test-hostname", "gQ3w7KzEFT543NdWZR-TVg"),
    ],
)
def test_hostname_header(mocker: MockerFixture, hash: bool, hostname: str, expected_result: str) -> None:
    mocker.patch("anaconda_auth.utils.gethostname", return_value=hostname)

    client = BaseClient(hash_hostname=hash)

    assert client.headers.get("X-Client-Hostname") == expected_result



from urllib.parse import urlparse
from urllib.parse import urlunparse

import pytest
from conda.cli.python_api import Commands
from conda.cli.python_api import run_command
from conda.gateways.connection.session import CondaHttpAuth
from conda.gateways.connection.session import CondaSession
from requests import HTTPError

from anaconda_auth._conda.repo_config import CondaTokenError
from anaconda_auth._conda.repo_config import get_ssl_verify
from anaconda_auth._conda.repo_config import token_list
from anaconda_auth._conda.repo_config import validate_token


def test_add_token(set_dummy_token, repodata_url, repo_url):
    assert token_list()[repo_url] == "SECRET"

    base_url = (
        repodata_url  # 'https://repo.anaconda.cloud/repo/main/osx-64/repodata.json'
    )
    scheme, netloc, path, *rest = urlparse(base_url)
    path = "/t/SECRET" + path
    token_url = urlunparse((scheme, netloc, path, *rest))
    assert CondaHttpAuth.add_binstar_token(base_url) == token_url


def test_channeldata_403(remove_token, channeldata_url):
    session = CondaSession()
    r = session.get(channeldata_url)
    with pytest.raises(HTTPError):
        r.raise_for_status()
    assert r.status_code == 403


@pytest.mark.skip(reason="blocking release in CI but passing fine locally")
def test_repodata_200(set_secret_token_mock_server, repodata_url):
    token_url = CondaHttpAuth.add_binstar_token(repodata_url)

    session = CondaSession()
    r = session.head(token_url)
    assert r.status_code == 200


# repo_url fixture configures test server, patches REPO_URL
def test_validate_token_error(repo_url):
    with pytest.raises(CondaTokenError):
        validate_token("SECRET")


# repo_url fixture configures test server, patches REPO_URL
@pytest.mark.skip(reason="blocking release in CI but passing fine locally")
def test_validate_token_works(secret_token, repo_url):
    assert validate_token(secret_token) is None


def test_conda_context():
    run_command(
        Commands.CONFIG, "--set", "ssl_verify", "false", use_exception_handler=True
    )
    assert not get_ssl_verify()



from __future__ import annotations

import sys
from typing import Generator

import pytest
from pytest import MonkeyPatch
from pytest_mock import MockerFixture

from anaconda_auth.cli import app
from anaconda_auth.client import BaseClient

from .conftest import CLIInvoker


@pytest.fixture
def is_a_tty(mocker: MockerFixture) -> Generator[None, None, None]:
    mocked = mocker.patch("anaconda_auth.cli.sys")
    mocked.stdout.isatty.return_value = True
    yield


@pytest.fixture
def is_not_a_tty(mocker: MockerFixture) -> Generator[None, None, None]:
    mocked = mocker.patch("anaconda_auth.cli.sys")
    mocked.stdout.isatty.return_value = False
    yield


@pytest.mark.usefixtures("disable_dot_env", "is_a_tty")
@pytest.mark.parametrize("subcommand", ["auth", "cloud"])
def test_login_required_tty(
    monkeypatch: MonkeyPatch,
    mocker: MockerFixture,
    invoke_cli: CLIInvoker,
    subcommand: str,
) -> None:
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)

    login = mocker.patch("anaconda_auth.cli.login")

    _ = invoke_cli([subcommand, "api-key"], input="n")
    login.assert_not_called()

    _ = invoke_cli([subcommand, "api-key"], input="y")
    login.assert_called_once()


@pytest.mark.usefixtures("disable_dot_env", "is_not_a_tty")
@pytest.mark.parametrize("subcommand", ["auth", "cloud"])
def test_login_error_handler_no_tty(
    monkeypatch: MonkeyPatch,
    mocker: MockerFixture,
    invoke_cli: CLIInvoker,
    subcommand: str,
) -> None:
    monkeypatch.delenv("ANACONDA_AUTH_API_KEY", raising=False)
    login = mocker.patch("anaconda_auth.cli.login")

    result = invoke_cli([subcommand, "api-key"])
    login.assert_not_called()

    assert "Login is required" in result.stdout


@pytest.mark.usefixtures("disable_dot_env")
@pytest.mark.parametrize("subcommand", ["auth", "cloud"])
def test_api_key_prefers_env_var(
    monkeypatch: MonkeyPatch, invoke_cli: CLIInvoker, subcommand: str
) -> None:
    monkeypatch.setenv("ANACONDA_AUTH_API_KEY", "foo")

    result = invoke_cli([subcommand, "api-key"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "foo"


@pytest.mark.usefixtures("disable_dot_env", "is_a_tty")
@pytest.mark.parametrize("subcommand", ["auth", "cloud"])
def test_http_error_login(
    monkeypatch: MonkeyPatch,
    invoke_cli: CLIInvoker,
    mocker: MockerFixture,
    subcommand: str,
) -> None:
    monkeypatch.setenv("ANACONDA_AUTH_API_KEY", "foo")
    login = mocker.patch("anaconda_auth.cli.login")

    result = invoke_cli([subcommand, "whoami"], input="y")
    login.assert_called_once()

    assert "is invalid" in result.stdout


@pytest.mark.usefixtures("is_a_tty")
@pytest.mark.parametrize("subcommand", ["auth", "cloud"])
def test_http_error_general(
    monkeypatch: MonkeyPatch,
    invoke_cli: CLIInvoker,
    mocker: MockerFixture,
    subcommand: str,
) -> None:
    @app.command("bad-request")
    def bad_request() -> None:
        client = BaseClient()
        res = client.get("api/docs/not-found")
        res.raise_for_status()

    result = invoke_cli([subcommand, "bad-request"])

    assert "404 Client Error" in result.stdout
    assert result.exit_code == 1


@pytest.mark.parametrize(
    "options",
    [
        ("-n", "someuser"),
        ("--name", "someuser"),
        ("-o", "someorg"),
        ("--org", "someorg"),
        ("--organization", "someorg"),
        ("--strength", "strong"),
        ("--strength", "weak"),
        ("--strong",),
        ("-w",),
        ("--weak",),
        ("--url", "https://some-server.com"),
        ("--max-age", "3600"),
        ("-s", "repo conda:download"),
        ("--scopes", "repo conda:download"),
        ("--out", "some-file.log"),
        ("-x",),
        ("--list-scopes",),
        ("-l",),
        ("--list",),
        ("-r", "token-1"),
        ("--remove", "token-1"),
        ("-c",),
        ("--create",),
        ("-i",),
        ("--info",),
        ("--current-info",),
    ],
)
def test_fallback_to_anaconda_client(
    options: tuple[str],
    invoke_cli: CLIInvoker,
    monkeypatch: MonkeyPatch,
    mocker: MockerFixture,
) -> None:
    """We fallback to anaconda-client for token management if any of its options are passed."""
    binstar_main = mocker.patch("binstar_client.scripts.cli.main")

    # Construct the CLI arguments
    args = ["auth", *options]

    # We need to override sys.argv since these get set by pytest
    monkeypatch.setattr(sys, "argv", ["some-anaconda-bin", *args])

    # Run the equivalent of `anaconda auth <options...>`
    result = invoke_cli(args)
    assert result.exit_code == 0

    # Calls are delegated to anaconda-client
    binstar_main.assert_called_once()
    binstar_main.assert_called_once_with(args, allow_plugin_main=False)



from pathlib import Path

import keyring
import pytest
from keyring.errors import PasswordDeleteError
from pytest import MonkeyPatch
from pytest_mock import MockerFixture

import anaconda_auth.token
from anaconda_auth.actions import logout
from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.config import AnacondaCloudConfig
from anaconda_auth.token import TokenExpiredError
from anaconda_auth.token import TokenInfo
from anaconda_auth.token import TokenNotFoundError


def test_expired_token_error(outdated_token_info: TokenInfo) -> None:
    with pytest.raises(TokenExpiredError):
        _ = outdated_token_info.get_access_token()


def test_token_not_found() -> None:
    config = AnacondaAuthConfig()

    with pytest.raises(TokenNotFoundError):
        _ = TokenInfo.load(config.domain)

    with pytest.raises(TokenNotFoundError):
        _ = TokenInfo(domain=config.domain).get_access_token()


def test_logout_multiple_okay(mocker: MockerFixture) -> None:
    """We can logout multiple times and no exception is raised."""
    import keyring

    delete_spy = mocker.spy(keyring, "delete_password")

    config = AnacondaAuthConfig(domain="test")
    token_info = TokenInfo(api_key="key", domain=config.domain)
    token_info.save()

    for _ in range(2):
        logout(config)

    delete_spy.assert_called_once()


def test_preferred_token_storage(monkeypatch: MonkeyPatch) -> None:
    import keyring.backend

    backends = {k.name: k for k in keyring.backend.get_all_keyring()}

    assert "token AnacondaKeyring" in backends
    assert backends["token AnacondaKeyring"].priority == 11.0
    assert (
        backends["token AnacondaKeyring"].priority
        > backends["chainer ChainerBackend"].priority
    )

    monkeypatch.setenv("ANACONDA_AUTH_PREFERRED_TOKEN_STORAGE", "system")
    backends = {k.name: k for k in keyring.backend.get_all_keyring()}

    assert "token AnacondaKeyring" in backends
    assert backends["token AnacondaKeyring"].priority == 0.2
    assert (
        backends["token AnacondaKeyring"].priority
        < backends["chainer ChainerBackend"].priority
    )


def test_anaconda_keyring_save_delete(tmp_path: Path) -> None:
    from anaconda_auth.token import AnacondaKeyring

    fn = tmp_path / "keyring"
    AnacondaKeyring.keyring_path = fn
    assert AnacondaKeyring.viable

    anaconda_keyring = AnacondaKeyring()

    assert anaconda_keyring.get_password("s", "u") is None

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u")

    anaconda_keyring.set_password("s", "u", "p")
    assert anaconda_keyring.keyring_path.exists()
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u": "p"}}'

    anaconda_keyring.set_password("s", "u2", "p")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u": "p", "u2": "p"}}'

    assert anaconda_keyring.viable

    assert anaconda_keyring.get_password("s", "u") == "p"
    assert anaconda_keyring.get_password("s", "u3") is None

    anaconda_keyring.delete_password("s", "u")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u2": "p"}}'

    anaconda_keyring.set_password("s2", "u", "p")
    assert (
        anaconda_keyring.keyring_path.read_text()
        == '{"s": {"u2": "p"}, "s2": {"u": "p"}}'
    )

    anaconda_keyring.delete_password("s", "u2")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {}, "s2": {"u": "p"}}'

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u2")

    assert anaconda_keyring.get_password("s3", "u4") is None


def test_anaconda_keyring_empty(tmp_path: Path) -> None:
    fn = tmp_path / "keyring"
    fn.touch()
    assert fn.exists()

    from anaconda_auth.token import AnacondaKeyring

    AnacondaKeyring.keyring_path = fn

    anaconda_keyring = AnacondaKeyring()
    assert anaconda_keyring.get_password("s", "u") is None

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u")


def test_anaconda_keyring_not_writable(tmp_path: Path) -> None:
    from anaconda_auth.token import AnacondaKeyring

    AnacondaKeyring.keyring_path = tmp_path / "keyring"
    AnacondaKeyring.keyring_path.touch()
    AnacondaKeyring.keyring_path.chmod(0x444)

    assert not AnacondaKeyring.viable


def test_anaconda_keyring_dir_not_a_dir(tmp_path: Path) -> None:
    from anaconda_auth.token import AnacondaKeyring

    keyring_dir = tmp_path / "anaconda"
    keyring_dir.touch()
    AnacondaKeyring.keyring_path = keyring_dir / "keyring"

    assert not AnacondaKeyring.viable


@pytest.fixture()
def expected_api_key() -> str:
    return "one key to rule them all"


@pytest.fixture(params=[("legacy-0", None), ("legacy-1", 1)])
def legacy_token_storage(request, expected_api_key, mocker: MockerFixture) -> str:
    """This fixture prepares the legacy token storage for two legacy domains, and the
    stored version associated with that domain. The assertions confirm the state of the
    keyring for each state.
    """
    mocker.patch.dict(
        anaconda_auth.token.MIGRATIONS, {"modern": ["legacy-1", "legacy-0"]}
    )

    legacy_domain, legacy_version = request.param

    # First make a token in the keyring with the legacy domain
    legacy_token = TokenInfo(
        api_key=expected_api_key, domain=legacy_domain, version=legacy_version
    )
    assert legacy_token.version is legacy_version
    legacy_token.save()

    payload = keyring.get_password(anaconda_auth.token.KEYRING_NAME, legacy_domain)
    assert payload is not None

    decoded = TokenInfo._decode(payload)
    if legacy_version is None:
        assert "version" not in decoded
    else:
        assert decoded["version"] == legacy_version

    payload = keyring.get_password(anaconda_auth.token.KEYRING_NAME, "modern")
    assert payload is None


def test_anaconda_keyring_domain_migration(
    expected_api_key: str, legacy_token_storage: None
) -> None:
    """Any of the legacy domains are migrated to the modern domain."""
    token = TokenInfo.load(domain="modern")
    assert token.api_key == expected_api_key
    assert token.version == 2

    payload = keyring.get_password(anaconda_auth.token.KEYRING_NAME, "legacy-0")
    assert payload is None

    payload = keyring.get_password(anaconda_auth.token.KEYRING_NAME, "legacy-1")
    assert payload is None

    payload = keyring.get_password(anaconda_auth.token.KEYRING_NAME, "modern")
    assert payload is not None

    decoded = TokenInfo._decode(payload)
    assert decoded["version"] == 2


def test_init_token_info_no_domain() -> None:
    """If we create a TokenInfo with no domain, it defaults to the config value."""
    config = AnacondaAuthConfig()
    token_info = TokenInfo()
    assert token_info.domain == config.domain


def test_load_token_info_create_false() -> None:
    with pytest.raises(TokenNotFoundError):
        _ = TokenInfo.load()


def test_load_token_info_create_true_config_domain() -> None:
    config = AnacondaAuthConfig()
    token_info = TokenInfo.load(create=True)
    assert token_info.domain == config.domain


def test_load_token_info_create_true_explicit_domain() -> None:
    expected_domain = "some-site.com"
    token_info = TokenInfo.load(domain=expected_domain, create=True)
    assert token_info.domain == expected_domain


def test_set_repo_token() -> None:
    token_info = TokenInfo()
    token_info.set_repo_token("org-name", "test-token")
    assert token_info.get_repo_token("org-name") == "test-token"


def test_set_repo_token_same_org_overwritten() -> None:
    token_info = TokenInfo()
    assert len(token_info.repo_tokens) == 0

    # Write the first token
    token_info.set_repo_token("org-name", "test-token")
    assert token_info.get_repo_token("org-name") == "test-token"
    assert len(token_info.repo_tokens) == 1

    # Token gets overwritten
    token_info.set_repo_token("org-name", "another-test-token")
    assert token_info.get_repo_token("org-name") == "another-test-token"
    assert len(token_info.repo_tokens) == 1


def test_set_repo_token_different_organization() -> None:
    token_info = TokenInfo()
    assert len(token_info.repo_tokens) == 0

    # Write the first token
    token_info.set_repo_token("org-name", "test-token")
    assert token_info.get_repo_token("org-name") == "test-token"
    assert len(token_info.repo_tokens) == 1

    # Write another token for a different organization
    token_info.set_repo_token("another-org-name", "another-test-token")
    assert token_info.get_repo_token("another-org-name") == "another-test-token"
    assert len(token_info.repo_tokens) == 2


def test_delete_repo_token() -> None:
    token_info = TokenInfo()
    assert len(token_info.repo_tokens) == 0

    # Write the first token
    token_info.set_repo_token("org-name", "test-token")
    assert token_info.get_repo_token("org-name") == "test-token"
    assert len(token_info.repo_tokens) == 1

    # Write another token for a different organization
    token_info.set_repo_token("another-org-name", "another-test-token")
    assert token_info.get_repo_token("another-org-name") == "another-test-token"
    assert len(token_info.repo_tokens) == 2

    # Delete the first token
    token_info.delete_repo_token("org-name")
    assert len(token_info.repo_tokens) == 1
    with pytest.raises(TokenNotFoundError):
        token_info.get_repo_token("org-name")


@pytest.mark.parametrize("saved_domain", ["anaconda.cloud", "anaconda.com"])
def test_token_reload_after_migration_via_anaconda_cloud_config(
    saved_domain: str,
) -> None:
    # Save the token (i.e. after login) to new and legacy domain in keyring
    original_token = TokenInfo(api_key="TEST_KEY", domain=saved_domain)
    original_token.save()

    # Now, load it via the explicit use of the AnacondaCloudConfig object with no
    # domain override. This is what the Assistant is doing. Since the token should
    # be migrated to anaconda.com when loading it if it is anaconda.cloud, this will
    # work when the AnacondaCloudConfig.domain default is anaconda.com.
    config = AnacondaCloudConfig()
    loaded_token = TokenInfo.load(domain=config.domain)
    assert loaded_token.api_key == "TEST_KEY"


@pytest.mark.parametrize(
    "saved_domains",
    [
        ["anaconda.cloud"],
        ["anaconda.com"],
        ["anaconda.cloud", "anaconda.com"],
    ],
)
def test_logout_removes_anaconda_cloud_tokens(saved_domains: str) -> None:
    # Given: The token is saved in keyring in either domain, or both
    for saved_domain in saved_domains:
        original_token = TokenInfo(api_key="TEST_KEY", domain=saved_domain)
        original_token.save()

    # When: I logout
    logout()

    # Then: neither of the domains contain a token
    for domain in ["anaconda.com", "anaconda.cloud"]:
        with pytest.raises(TokenNotFoundError):
            TokenInfo.load(domain)


def test_anaconda_keyring_name():
    from anaconda_auth.token import AnacondaKeyring
    assert AnacondaKeyring.name == "token AnacondaKeyring"



import warnings
from functools import cached_property
from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Union
from urllib.parse import urljoin

import requests
from pydantic import BaseModel
from pydantic_settings import SettingsConfigDict

from anaconda_auth import __version__ as version
from anaconda_cli_base.config import AnacondaBaseSettings
from anaconda_cli_base.console import console


def _raise_deprecated_field_set_warning(set_fields: Dict[str, Any]) -> None:
    fields_str = ", ".join(sorted(f'"{s}"' for s in set_fields.keys()))
    warning_text = (
        "The following fields have been set using legacy environment variables "
        + "prefixed with 'ANACONDA_CLOUD_` or in the `plugins.cloud` section "
        + f"of `~/.anaconda/config.toml`: {fields_str}\n\n"
        + "Please either rename environment variables to the corresponding "
        + "`ANACONDA_AUTH_` version, or replace the `plugins.cloud` section "
        + "of the config file with `plugins.auth`."
    )
    console.print(f"[red]{warning_text}[/red]")
    warnings.warn(
        warning_text,
        DeprecationWarning,
    )


class AnacondaAuthConfig(AnacondaBaseSettings, plugin_name="auth"):
    preferred_token_storage: Literal["system", "anaconda-keyring"] = "anaconda-keyring"
    domain: str = "anaconda.com"
    auth_domain_override: Optional[str] = None
    api_key: Optional[str] = None
    ssl_verify: bool = True
    extra_headers: Optional[Union[Dict[str, str], str]] = None
    client_id: str = "b4ad7f1d-c784-46b5-a9fe-106e50441f5a"
    redirect_uri: str = "http://127.0.0.1:8000/auth/oidc"
    openid_config_path: str = "/.well-known/openid-configuration"
    oidc_request_headers: Dict[str, str] = {"User-Agent": f"anaconda-auth/{version}"}
    login_success_path: str = "/app/local-login-success"
    login_error_path: str = "/app/local-login-error"
    use_unified_repo_api_key: bool = False
    hash_hostname: bool = True

    def __init__(self, **kwargs: Any):
        if self.__class__ == AnacondaAuthConfig:
            config = AnacondaCloudConfig(raise_deprecation_warning=False)
            set_fields = config.model_dump(exclude_unset=True)
            if set_fields:
                _raise_deprecated_field_set_warning(set_fields)

                # Merge dictionaries, ensuring that any duplicate keys in kwargs wins
                kwargs = {**set_fields, **kwargs}
        super().__init__(**kwargs)

    @property
    def auth_domain(self) -> str:
        """The authentication domain base URL.

        Defaults to the `auth` subdomain of the main domain.

        """
        if self.auth_domain_override:
            return self.auth_domain_override
        return f"auth.{self.domain}"

    @property
    def well_known_url(self) -> str:
        """The URL from which to load the OpenID configuration."""
        return urljoin(f"https://{self.auth_domain}", self.openid_config_path)

    @property
    def login_success_url(self) -> str:
        """The location to redirect after auth flow, if successful."""
        return urljoin(f"https://{self.domain}", self.login_success_path)

    @property
    def login_error_url(self) -> str:
        """The location to redirect after auth flow, if there is an error."""
        return urljoin(f"https://{self.domain}", self.login_error_path)

    @property
    def oidc(self) -> "OpenIDConfiguration":
        """The OIDC configuration, cached as a regular instance attribute."""
        res = requests.get(
            self.well_known_url,
            headers=self.oidc_request_headers,
            verify=self.ssl_verify,
        )
        res.raise_for_status()
        oidc_config = OpenIDConfiguration(**res.json())
        return self.__dict__.setdefault("_oidc", oidc_config)

    @cached_property
    def aau_token(self) -> Union[str, None]:
        # The token is cached in anaconda_anon_usage, so we can also cache here
        try:
            from anaconda_anon_usage.tokens import token_string
        except ImportError:
            return None

        try:
            return token_string()
        except Exception:
            # We don't want this to block user login in any case,
            # so let any Exceptions pass silently.
            return None


class OpenIDConfiguration(BaseModel):
    authorization_endpoint: str
    token_endpoint: str


_OLD_OIDC_REQUEST_HEADERS = {"User-Agent": f"anaconda-cloud-auth/{version}"}


class AnacondaCloudConfig(AnacondaAuthConfig, plugin_name="cloud"):
    # Here, we explicitly specify the model_config for this class. This is because
    # there is a bug inside AnacondaBaseSettings, where the env_prefix is mutated
    # in that base class. Thus, nested inheritance doesn't quite work as I'd expect.
    # However, if we set this attribute on *this* class, then that problem goes away,
    # Even though the behavior that handles the injecting of the `plugin_name` into
    # the env_prefix is handled in the __init_subclass__ method in that base class.
    model_config = SettingsConfigDict(
        env_file=".env",
        pyproject_toml_table_header=(),
        env_prefix="ANACONDA_",
        env_nested_delimiter="__",
        extra="ignore",
        ignored_types=(cached_property,),
    )
    oidc_request_headers: Dict[str, str] = _OLD_OIDC_REQUEST_HEADERS

    def __init__(self, raise_deprecation_warning: bool = True, **kwargs: Any):
        if raise_deprecation_warning:
            warnings.warn(
                "AnacondaCloudConfig is deprecated, please use AnacondaAuthConfig instead.",
                DeprecationWarning,
            )

        super().__init__(**kwargs)



import pytest
from requests import PreparedRequest
from requests import Response
from requests.hooks import dispatch_hook

from anaconda_auth.token import TokenInfo

conda = pytest.importorskip("conda")

from conda.gateways.connection.session import CondaSession  # noqa: E402
from conda.gateways.connection.session import get_session  # noqa: E402

from anaconda_auth._conda.auth_handler import AnacondaAuthError  # noqa: E402
from anaconda_auth._conda.auth_handler import AnacondaAuthHandler  # noqa: E402


@pytest.fixture()
def mocked_empty_conda_token(mocker):
    mocker.patch(
        "anaconda_auth._conda.repo_config.token_list",
        return_value={},
    )


@pytest.fixture()
def mocked_conda_token(mocker):
    mocker.patch(
        "anaconda_auth._conda.repo_config.token_list",
        return_value={"https://repo.anaconda.cloud/repo/": "my-test-token"},
    )


@pytest.fixture()
def mocked_token_info(mocker):
    mocker.patch(
        "anaconda_auth.token.TokenInfo.load",
        return_value=TokenInfo(
            domain="repo.anaconda.cloud",
            repo_tokens=[
                {
                    "org_name": "my-first-org",
                    "token": "my-first-test-token-in-token-info",
                },
                {"org_name": "my-org", "token": "my-test-token-in-token-info"},
            ],
        ),
    )


@pytest.fixture()
def mocked_token_info_with_api_key(mocker):
    mocker.patch(
        "anaconda_auth.token.TokenInfo.load",
        return_value=TokenInfo(
            domain="repo.anaconda.cloud",
            api_key="my-test-api-key",
            repo_tokens=[
                {
                    "org_name": "my-first-org",
                    "token": "my-first-test-token-in-token-info",
                },
                {"org_name": "my-org", "token": "my-test-token-in-token-info"},
            ],
        ),
    )


@pytest.fixture()
def handler():
    return AnacondaAuthHandler(
        channel_name="https://repo.anaconda.cloud/repo/my-org/my-channel"
    )


@pytest.mark.usefixtures("mocked_conda_token")
def test_get_token_via_conda_token(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-token"


@pytest.mark.usefixtures("mocked_token_info")
def test_get_repo_token_via_keyring(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-token-in-token-info"


@pytest.mark.usefixtures("mocked_token_info_with_api_key")
def test_get_unified_api_token_via_keyring(handler, monkeypatch):
    monkeypatch.setenv("ANACONDA_AUTH_USE_UNIFIED_REPO_API_KEY", "True")
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-api-key"


@pytest.mark.usefixtures("mocked_token_info")
def test_auth_handler_call_sets_authorization_header_repo_token(handler, monkeypatch):
    request = PreparedRequest()
    request.url = (
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    request.headers = {}

    modified_request = handler(request)

    assert (
        modified_request.headers["Authorization"] == "token my-test-token-in-token-info"
    )


@pytest.mark.usefixtures("mocked_token_info_with_api_key")
def test_auth_handler_call_sets_authorization_header_unified_api_token(
    handler, monkeypatch
):
    monkeypatch.setenv("ANACONDA_AUTH_USE_UNIFIED_REPO_API_KEY", "True")

    request = PreparedRequest()
    request.url = (
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    request.headers = {}

    modified_request = handler(request)

    assert modified_request.headers["Authorization"] == "Bearer my-test-api-key"


@pytest.mark.usefixtures("mocked_token_info")
def test_get_token_for_main_finds_first_token(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/main/noarch/repodata.json"
    )
    assert token == "my-first-test-token-in-token-info"


@pytest.mark.usefixtures("mocked_empty_conda_token")
def test_get_token_missing(handler):
    with pytest.raises(AnacondaAuthError):
        _ = handler._load_token(
            "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
        )


@pytest.fixture()
def url() -> str:
    return "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"


@pytest.fixture()
def session(handler, url) -> CondaSession:
    # Create a session and assign the handler to it
    get_session.cache_clear()
    session_obj = get_session(url)
    session_obj.auth = handler
    return session_obj


@pytest.mark.usefixtures("mocked_token_info")
def test_inject_header_during_request(session, url, monkeypatch):
    # Set up a dummy function that will capture the PreparedRequest without sending it.
    request = None

    def capture_request(req, *args, **kwargs):
        nonlocal request
        request = req

    monkeypatch.setattr(session, "send", capture_request)

    # Make sure the token got injected
    session.get(url)
    assert request.headers.get("Authorization") == "token my-test-token-in-token-info"


@pytest.mark.usefixtures("mocked_token_info")
def test_response_callback_403(session, url, monkeypatch):
    def trigger_403(req, *args, **kwargs):
        response = Response()
        response.status_code = 403
        response = dispatch_hook("response", req.hooks, response, **kwargs)
        return response

    monkeypatch.setattr(session, "send", trigger_403)

    # A 403 response is captured by the hook and a custom exception is raised
    with pytest.raises(AnacondaAuthError):
        session.get(url)



import os
import warnings
from unittest import mock

import pytest

conda = pytest.importorskip("conda")

from conda.cli.python_api import Commands  # noqa: E402
from conda.cli.python_api import run_command  # noqa: E402

from anaconda_auth._conda.repo_config import clean_index  # noqa: E402
from anaconda_auth._conda.repo_config import token_remove  # noqa: E402
from anaconda_auth._conda.repo_config import token_set  # noqa: E402


def pytest_configure(config):
    warnings.filterwarnings("always")


@pytest.fixture(scope="session")
def test_server_url() -> str:
    """Run a test server, and return its URL."""
    from . import testing_server

    return testing_server.run_server()


@pytest.fixture
def repo_url(test_server_url: str) -> str:
    repo_url = test_server_url + "/repo/"
    with mock.patch.dict(os.environ, {"CONDA_TOKEN_REPO_URL": repo_url}):
        with mock.patch("anaconda_auth._conda.repo_config.REPO_URL", repo_url):
            yield repo_url


@pytest.fixture(scope="session", autouse=True)
def reset_channels_alias():
    clean_index()
    run_command(Commands.CONFIG, "--remove-key", "channels", use_exception_handler=True)
    run_command(
        Commands.CONFIG, "--prepend", "channels", "defaults", use_exception_handler=True
    )
    run_command(
        Commands.CONFIG,
        "--set",
        "channel_alias",
        "https://conda.anaconda.org",
        use_exception_handler=True,
    )


@pytest.fixture(scope="function", autouse=True)
def set_ssl_verify_true():
    run_command(
        Commands.CONFIG, "--set", "ssl_verify", "true", use_exception_handler=True
    )
    yield
    run_command(
        Commands.CONFIG, "--set", "ssl_verify", "true", use_exception_handler=True
    )


@pytest.fixture(scope="function")
def remove_token(repo_url):
    token_remove()
    yield
    token_remove()


@pytest.fixture(scope="session", autouse=True)
def remove_token_end_of_session():
    yield
    token_remove()


@pytest.fixture(scope="function")
def remove_token_no_repo_url_mock():
    """
    Remove token without mock repo_url
    """
    token_remove()
    yield
    token_remove()


@pytest.fixture(scope="function")
def set_dummy_token(repo_url):
    token_remove()
    token_set("SECRET", force=True)
    yield
    token_remove()


@pytest.fixture(scope="function")
def set_secret_token():
    token_remove()
    secret_token = os.environ.get("CE_TOKEN", "SECRET_TOKEN")
    token_set(secret_token, force=True)
    yield
    token_remove()


@pytest.fixture(scope="function")
def set_secret_token_mock_server(repo_url):
    token_remove()
    secret_token = os.environ.get("CE_TOKEN", "SECRET_TOKEN")
    token_set(secret_token, force=True)
    yield
    token_remove()


@pytest.fixture(scope="function")
def set_secret_token_with_signing():
    token_remove()
    secret_token = os.environ.get("CE_TOKEN", "SECRET_TOKEN")
    token_set(secret_token, enable_signature_verification=True, force=True)
    yield
    token_remove()


@pytest.fixture(scope="function")
def secret_token():
    token = os.environ.get("CE_TOKEN", "SECRET_TOKEN")
    yield token


@pytest.fixture(scope="function")
def uninstall_rope():
    run_command(Commands.REMOVE, "rope", "-y", "--force", use_exception_handler=True)
    yield
    run_command(Commands.REMOVE, "rope", "-y", "--force", use_exception_handler=True)


@pytest.fixture
def channeldata_url(repo_url):
    return repo_url + "main/channeldata.json"


@pytest.fixture
def repodata_url(repo_url):
    return repo_url + "main/osx-64/repodata.json"



import json
from textwrap import dedent

import pytest
from conda.cli.python_api import Commands
from conda.cli.python_api import run_command
from packaging.version import parse

from anaconda_auth._conda.repo_config import CONDA_VERSION


def json_skip_preamble(text):
    """
    Ignore text before the first line starting with {
    """
    capture = False
    captured = []
    for line in text.splitlines():
        if line.strip().startswith(("{", "[")):
            capture = True
        if capture:
            captured.append(line)
    return json.loads("\n".join(captured))


def test_skip_garbage():
    lines = dedent(
        """
        Random text
        Error message
         {"foo":
        "bar"
        }
        """
    )

    lines2 = dedent(
        """
        Unexpected warning
         [ 1 ]
        """
    )
    assert json_skip_preamble(lines) == {"foo": "bar"}
    assert json_skip_preamble(lines2) == [1]


@pytest.mark.skipif(
    CONDA_VERSION < parse("4.10.1"),
    reason="Signature verification was added in Conda 4.10.1",
)
@pytest.mark.skipif(
    parse("23.9") <= CONDA_VERSION < parse("23.10"),
    reason="metadata_signature_status missing in conda 23.9",
)
@pytest.mark.skipif(
    CONDA_VERSION >= parse("24.1"),
    reason="conda >=24.1 delays signature checks to after a solve",
)
def test_conda_search_rope_signed(set_secret_token_with_signing):
    stdout, _, _ = run_command(Commands.SEARCH, "--spec", "rope=0.18.0=py_0", "--json")
    try:
        rope = json_skip_preamble(stdout)["rope"][0]
    except json.JSONDecodeError:
        print("Could not decode", stdout)
        raise
    # conda 23.x appears to have changed this value to a string
    assert rope.get("metadata_signature_status") in (
        0,
        "(INFO: package metadata is signed by Anaconda and trusted)",
    ), rope

    stdout, _, _ = run_command(
        Commands.SEARCH, "--spec", "conda-forge::rope=0.18.0=pyhd3deb0d_0", "--json"
    )
    try:
        rope = json_skip_preamble(stdout)["rope"][0]
    except json.JSONDecodeError:
        print("Could not decode", stdout)
        raise
    # conda 23.x appears to omit the key if unsigned
    assert "metadata_signature_status" not in rope or rope.get(
        "metadata_signature_status"
    ) in (
        -1,
        "(WARNING: metadata signature verification failed)",
    ), rope


@pytest.mark.integration
def test_conda_search_rope(set_secret_token):
    if CONDA_VERSION < parse("4.4"):
        stdout, _, _ = run_command(
            Commands.SEARCH, "--spec", "rope=0.18.0=py_0", "--json"
        )
    else:
        stdout, _, _ = run_command(Commands.SEARCH, "rope==0.18.0=py_0", "--json")
    try:
        rope = json_skip_preamble(stdout)["rope"][0]
    except json.JSONDecodeError:
        print("Could not decode", stdout)
        raise
    assert rope["url"].startswith("https://repo.anaconda.cloud/repo/main/noarch")


@pytest.mark.integration
def test_conda_install_rope(set_secret_token, uninstall_rope):
    install_args = (Commands.INSTALL, "rope", "-y")
    # is libmamba-solver using its own network code, not adding the token here?
    if parse("23.10") < CONDA_VERSION < parse("24.1"):
        install_args = install_args + ("--solver=classic",)
    run_command(*install_args)

    if CONDA_VERSION < parse("4.6"):
        stdout, _, _ = run_command(Commands.LIST, "--explicit")
        pkgs = stdout.splitlines()
        for p in pkgs:
            if "rope" in p:
                assert p.startswith("https://repo.anaconda.cloud/repo/main")
    else:
        stdout, _, _ = run_command(
            Commands.LIST, "rope", "--show-channel-urls", "--json"
        )
        try:
            rope = json_skip_preamble(stdout)[0]
        except (json.JSONDecodeError, TypeError):
            print("Could not decode", stdout)
            raise
        assert rope["base_url"] == "https://repo.anaconda.cloud/repo/main"


@pytest.mark.integration
def test_conda_install_with_conda_forge(set_secret_token, uninstall_rope):
    run_command(
        Commands.INSTALL,
        "-c",
        "defaults",
        "-c",
        "conda-forge",
        "rope",
        "conda-forge-pinning",
        "-y",
    )

    if CONDA_VERSION < parse("4.6"):
        stdout, _, _ = run_command(Commands.LIST, "--explicit")
        pkgs = stdout.splitlines()
        for p in pkgs:
            if "rope" in p:
                assert p.startswith("https://repo.anaconda.cloud/repo/main")
            if "conda-forge-pinning" in p:
                assert p.startswith("https://conda.anaconda.org/conda-forge")
    else:
        stdout, _, _ = run_command(
            Commands.LIST, "rope", "--show-channel-urls", "--json"
        )
        try:
            rope = json_skip_preamble(stdout)[0]
        except json.JSONDecodeError:
            print("Could not decode", stdout)
            raise
        assert rope["base_url"] == "https://repo.anaconda.cloud/repo/main"

        stdout, _, _ = run_command(
            Commands.LIST, "conda-forge-pinning", "--show-channel-urls", "--json"
        )
        try:
            conda_forge_pinning = json_skip_preamble(stdout)[0]
        except json.JSONDecodeError:
            print("Could not decode", stdout)
            raise
        assert (
            conda_forge_pinning["base_url"] == "https://conda.anaconda.org/conda-forge"
        )
